
"""
This module generates
	metfunc
objects from templates.
"""

import numpy as np


from xhorizon.tortoise.metfunc_class import metfunc
from xhorizon.tortoise import tortoise
from xhorizon.tortoise import math_util


"""
Templates.
Each template calls build_metfunc(params) in order to build the object from specified parameters.
"""

def minkowski():
	############################# INPUT #############################
	## parameters
	fparams = dict()
	Fparams = dict(eps=1e-9, inf=25., npoints_interp=500)
	## metric function
	f = lambda r: 1. + 0.*r
	## zeroes and slopes of f(r)
	ri = np.array([])
	ki = np.array([])
	## info
	info = {
		'Type' : "Minkowski",
		'Metric Function' : r"$f(r) = 1$",
		'Parameters' : '',
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func





def schwarzschild(R=1.):
	############################# INPUT #############################
	## parameters
	R = float(R)
	fparams = dict(R=R)
	Fparams = dict(eps=1e-6*R, inf=25., npoints_interp=500)
	## metric function
	f = lambda r: 1. - R / r
	## zeroes and slopes of f(r)
	ri = np.array([R])
	ki = np.array([1./R])
	## info
	info = {
		'Type' : "Schwarzschild",
		'Metric Function' : r"$f(r) = 1 - R/r$",
		'Parameters' : ', '.join([r"$R=%r$"%R])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


def reissner_nordstrom(M=0.5, Q=0.3):
	############################# INPUT #############################
	## parameters
	M, Q = float(M), float(Q)
	fparams = dict(M=M,Q=Q)
	Fparams = dict(eps=1e-9, inf=25., npoints_interp=500)
	## metric function
	f = lambda r: 1. - 2.*M/r + Q**2/r**2
	## zeroes and slopes of f(r)
	ri = math_util.pos_real_roots(np.array([1., -2.*M, Q**2]))
	ki = 2. * ri**(-3) * ( M*ri - Q**2)
	## info
	info = {
		'Type' : "Reissner Nordstrom",
		'Metric Function' : r"$f(r) = 1 - 2M/r + Q^2/r^2$",
		'Parameters' : ', '.join([r"$M=%r$"%M,r"$Q=%r$"%Q])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


def hayward(R=1.,l=0.1):
	############################# INPUT #############################
	## parameters
	R, l = float(R), float(l)
	fparams = dict(R=R,l=l)
	Fparams = dict(eps=1e-9, inf=25., npoints_interp=500)
	## metric function
	f = lambda r: 1. - R*r**2 / ( R*l**2 + r**3 )
	## zeroes and slopes of f(r)
	ri = math_util.pos_real_roots(np.array([1.,-R,0.,R*l**2]))
	ki = ( R * ri**4 - 2. * R**2 * l**2 * ri ) / ( R * l**2 + ri**3 )**2
	## info
	info = {
		'Type' : "Hayward",
		'Metric Function' : r"$f(r) = 1 - R r^2 / (R l^2 + r^3)$",
		'Parameters' : ', '.join([r"$R=%r$"%R,r"$l=%r$"%l])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


def dS(L=10.):
	############################# INPUT #############################
	## parameters
	L = float(L)
	fparams = dict(L=L)
	Fparams = dict(eps=1e-9, inf=20.*L, npoints_interp=500)
	## metric function
	f = lambda r: 1. - (r/L)**2
	## zeroes and slopes of f(r)
	ri = np.array([L])
	ki = np.array([-2./L])
	## info
	info = {
		'Type' : "de Sitter",
		'Metric Function' : r"$f(r) = 1 - (r/L)^2$",
		'Parameters' : r"$L=%r$"%L
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


def AdS(L=10.):
	############################# INPUT #############################
	## parameters
	L = float(L)
	fparams = dict(L=L)
	Fparams = dict(eps=1e-9, inf=20.*L, npoints_interp=500)
	## metric function
	f = lambda r: 1. + (r/L)**2
	## zeroes and slopes of f(r)
	ri = np.array([])
	ki = np.array([])
	## info
	info = {
		'Type' : "Anti de Sitter",
		'Metric Function' : r"$f(r) = 1 + (r/L)^2$",
		'Parameters' : r"$L=%r$"%L
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func



def Hay_dS(l=.1, R=1., L=10.):
	############################# INPUT #############################
	## parameters
	l, R, L = float(l), float(R), float(L)
	fparams = dict(l=l,R=R,L=L)
	Fparams = dict(eps=1e-9, inf=20.*L, npoints_interp=500)
	## metric function
	f = lambda r: 1. - R*r**2/(R*l**2+r**3) - r**2/L**2
	## zeroes and slopes of f(r)
	ri = math_util.pos_real_roots(np.array([-1.,0.,L**2,-R*l**2-R*L**2,0.,R*(L**2)*(l**2)]))
	ki = ( R * ri**4 - 2. * R**2 * l**2 * ri ) / ( R * l**2 + ri**3 )**2 - 2.*ri / L**2
	## info
	info = {
		'Type' : "Hayward - de Sitter",
		'Metric Function' : r"$f(r) = 1 - Rr^2/(Rl^2+r^3) - (r/L)^2$",
		'Parameters' : ', '.join([r"$l=%r$"%l,r"$R=%r$"%R,r"$L=%r$"%L])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


def Hay_AdS(l=.1, R=1., L=10.):
	############################# INPUT #############################
	## parameters
	l, R, L = float(l), float(R), float(L)
	fparams = dict(l=l,R=R,L=L)
	Fparams = dict(eps=1e-9, inf=20.*L, npoints_interp=500)
	## metric function
	f = lambda r: 1. - R*r**2/(R*l**2+r**3) + r**2/L**2
	## zeroes and slopes of f(r)
	ri = math_util.pos_real_roots(np.array([ 1.,0.,L**2, R*l**2-R*L**2,0.,R*(L**2)*(l**2)]))
	ki = ( R * ri**4 - 2. * R**2 * l**2 * ri ) / ( R * l**2 + ri**3 )**2 + 2.*ri / L**2
	## info
	info = {
		'Type' : "Hayward - Anti de Sitter",
		'Metric Function' : r"$f(r) = 1 - Rr^2/(Rl^2+r^3) + (r/L)^2$",
		'Parameters' : ', '.join([r"$l=%r$"%l,r"$R=%r$"%R,r"$L=%r$"%L])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


def Hay_dS_cutoff(l=.1, R=1., L=10.):
	############################# INPUT #############################
	## parameters
	l, R, L = float(l), float(R), float(L)
	fparams = dict(l=l,R=R,L=L)
	Fparams = dict(eps=1e-9, inf=20.*L, npoints_interp=500)
	## L0
	ri = math_util.pos_real_roots(np.array([-1.,0.,L**2,-R*l**2-R*L**2,0.,R*(L**2)*(l**2)]))
	L0 = ri[-1] - 0.1*(ri[-1] - ri[-2])
	## metric function
	def f(r):
		if type(r) in [float, np.float64]:
			if r<=L0:
				return 1. - r**2/L**2
			if r>L0:
				return 1. - R*r**2/(R*l**2+r**3) - r**2/L**2
		if type(r)==np.ndarray:
			ff = np.nan*r
			mask = r> L0
			ff[mask] = 1. - R*r[mask]**2/(R*l**2+r[mask]**3) - r[mask]**2/L**2
			mask = r<=L0
			ff[mask] = 1. - r[mask]**2/L**2
			return 1.*ff
	## zeroes and slopes of f(r)
	ri = math_util.pos_real_roots(np.array([-1.,0.,L**2,-R*l**2-R*L**2,0.,R*(L**2)*(l**2)]))
	ki = ( R * ri**4 - 2. * R**2 * l**2 * ri ) / ( R * l**2 + ri**3 )**2 - 2.*ri / L**2
	## should be just one though because of cutoff
	ri = ri[-1:]
	ki = ki[-1:]
	## info
	info = {
		'Type' : "Hayward - de Sitter with Cutoff",
		'Metric Function' : r"De Sitter inside cutoff length, Hay-dS outside cutoff length.",
		'Parameters' : ', '.join([r"$l=%r$"%l,r"$R=%r$"%R,r"$L=%r$"%L])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func



def negativeAdS(L=1.):
	############################# INPUT #############################
	## parameters
	L = float(L)
	fparams = dict(L=L)
	Fparams = dict(eps=1e-9, inf=50., npoints_interp=500)
	## metric function
	f = lambda r: -( 1. + (r/L)**2 )
	## zeroes and slopes of f(r)
	ri = np.array([])
	ki = np.array([])
	## info
	info = {
		'Type' : "Negative Anti de Sitter",
		'Metric Function' : r"$f(r) = -(1 + (r/L)^2)$",
		'Parameters' : r"$L=%r$"%L
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func



def schwarzschild_dS(R=1.,L=10.):
	############################# INPUT #############################
	## parameters
	R, L = float(R), float(L)
	fparams = dict(R=R,L=L)
	Fparams = dict(eps=1e-9, inf=25., npoints_interp=500)
	## metric function
	f = lambda r: 1. - R/r - r**2/L**2
	## zeroes and slopes of f(r)
	ri = math_util.pos_real_roots(np.array([1.,0.,-(L**2),R*L**2]))
	ki = R/ri**2 - 2.*ri/L**2
	## info
	info = {
		'Type' : "Schwarzschild - de Sitter",
		'Metric Function' : r"$f(r) = 1 - R/r - (r/L)^2$",
		'Parameters' : ', '.join([r"$R=%r$"%R,r"$L=%r$"%L])
		}
	#################################################################
	## store params in dict
	params = dict(fparams=fparams, Fparams=Fparams, f=f, ri=ri, ki=ki, info=info)
	## build
	func = build_metfunc(params)
	## return
	return func


"""
Generic build from params.
"""

def build_metfunc(params):
	## get build parameters from input dictionary
	fparams = params['fparams']
	Fparams = params['Fparams']
	f       = params['f']
	ri      = params['ri']
	ki      = params['ki']
	info    = params['info']
	## create object
	func = metfunc()
	## fill fields using specified values
	func.f = f
	func.fparams = fparams
	func.Fparams = Fparams
	func.rj = np.concatenate( [ np.array([0.]), ri, np.array([Fparams['inf']]) ] )
	func.kj = np.concatenate( [ np.array([0.]), ki, np.array([0.]) ] )
	func.info = info
	## generate tortoise function and inverse
	print("\nThis part is slow, reference F(r) arrays being calculated by integration...")
	tort_by_interp = tortoise.F_by_interp(func.f, func.rj, func.Fparams['eps'], npoints=func.Fparams['npoints_interp'])
	print("...done.\n")
	func.F = tort_by_interp.make_F()
	func.Finv = tort_by_interp.make_Finv()
	## define interval getter, first trapped interval
	func.winterval = lambda r,asint=True: tortoise.which_interval(r, func.rj, asint)
	func.first_trapped_j = tortoise.first_trapped_j(func.f,func.rj)
	## define interpolation reference values
	func.r_ref = tort_by_interp.r_ref
	func.rstar_ref = tort_by_interp.rstar_ref
	## return
	return func










"""
Tests.
Tests run if __name__='__main__'.
"""


def test1():
	"""
	"""
	######### input ########
	func  = hayward(R=.05, l=.01)
	########################
	print("\nTEST 9")
	print(func)
	print(func.rj)
	print(func.info['Type'])
	##
	import matplotlib.pyplot as plt
	plt.plot(func.r_ref, func.rstar_ref, 'kx')
	plt.xlim(0,25)
	plt.grid()
	plt.show()
	## end
	print("END TEST 9\n")



## run tests if __name__="__main__"

if __name__=="__main__":
	print("RUN TESTS")
	test1()





def demo():
	import numpy as np
	import matplotlib.pyplot as plt
	import xhorizon as xh
	from xhorizon.evap.massplot import massplot

	"""
	"""

	## setup figure #################
	## rc
	plt.rcParams['font.size'] = 7
	plt.rcParams['axes.linewidth'] = .4
	plt.rcParams['xtick.major.size'] = 2
	plt.rcParams['ytick.major.size'] = 2
	plt.rcParams['lines.solid_capstyle'] = 'butt'
	plt.rcParams['lines.dash_capstyle'] = 'butt'
	## figure size
	fig_width  = 2.5  ## inches
	fig_aspect = 1.68   ## height/width
	fig_height = fig_width * fig_aspect
	## axis size
	width  = 0.86   ## x fraction
	aspect = 1.84   ## absolute height/width
	height = aspect * width / fig_aspect    ## y fraction
	## axis loc
	left, bottom = .12, .05     ## x fraction, y fraction
	## define figure and axes
	plt.figure(1,figsize=(fig_width,fig_height))
	ax1 = plt.axes([left,bottom,width,height], aspect=1.)
	ax = [ax1]
	## format axes
	for axx in ax:
		## set axis
		plt.sca(axx)
		## labels
		plt.xlabel('$V-U$', labelpad=-7)
		plt.ylabel('$V+U$', labelpad=-4)
		## ticks
		plt.xticks([1,2])
		plt.yticks([-1,0,1])
		## lims
		sq = 1.2
		x0 = .86
		plt.xlim(x0,x0+sq)
		plt.ylim(-0.5*aspect*sq,0.5*aspect*sq)
	################################
	

	## save
	path = "temp-figs/demo"
	sfp = dict(dpi=800)
	temp_only = False

	## draw
	draw = True

	## label
	label = '$(a)$'

	# func type
	ftype = 2

	## input	
	L = 1.
	l  = .01
	le = .01

	## input
	R  = .2

	## evap
	Nevap = 1
	Tevap = 10.

	## accrete
	Nacc = 1
	Tacc = .75

	## seed
	seed = 0

	## vv
	voff = 0. #-Tacc - 0.
	veta = 1.

	## uu
	uoff = 0.
	ueta = 0.


	## params
	params = dict()
	## funcs
	#### schwarz
	if ftype==0:
		params.update(dict(functype0=xh.mf.minkowski, fparams0=dict(), functype1=xh.mf.schwarzschild, fparams1=dict()))
	#### hayward
	if ftype==1:
		params.update(dict(functype0=xh.mf.minkowski, fparams0=dict(), functype1=xh.mf.hayward, fparams1=dict(l=1.*l)))
	#### AdS
	if ftype==2:
		params.update(dict(functype0=xh.mf.AdS, fparams0=dict(L=1.*L), functype1=xh.mf.Hay_AdS, fparams1=dict(l=1.*l, L=1.*L)))
	#### AdS
	if ftype==3:
		params.update(dict(functype0=xh.mf.dS, fparams0=dict(L=1.*L), functype1=xh.mf.Hay_dS, fparams1=dict(l=1.*l, L=1.*L)))
	## evap
	params.update(dict(Rmax=1.*R, le=1.*le, Tevap=1.*Tevap, Nevap=1*Nevap))
	## accrete
	params.update(dict(Tacc=1.*Tacc, Naccrete=1*Nacc))
	## offset
	params.update(dict(voff=1.*voff, veta=1.*veta, uoff=1.*uoff, ueta=1.*ueta))


##########################################
	## change lim if seed
	if seed not in [0,-1]:
		## format axes
		for axx in ax:
			## set axis
			plt.sca(axx)
			## ticks
			plt.xticks([-1,0,1,2,3])
			plt.yticks([-1,0,1,2,3])
			## lims
			sq = 3.4
			x0, y0 = -1.2, 1.2
			plt.xlim(x0,x0+sq)
			plt.ylim(y0-0.5*aspect*sq,y0+0.5*aspect*sq)
##########################################


	## go
	reglist, chainparams = xh.evap.create_evap(params, seed=seed)

	## draw
	if draw:
		print("plot")
		pp = dict(l=1.*l, R=1.*R)
		xh.evap.drawreg(reglist, chainparams, fparams=pp)


	## label
	if True:
		plt.annotate(s=label, xy=(.95,.97), xycoords='axes fraction', ha='right', va='top', size=8)


	## auto param label
	if True:
		plabel = [r"$l_{ev}=%3s$"%(le), r"", r"$l=%3s$"%(l), r"$2M=%3s$"%(R), r"$L=%3s$"%(L)]
		#plabel += [r"", r"$\tau_{acc}=%3s$"%(Tacc), r"$\tau_{ev}=%3s$"%(Tevap)]
		plabel = "\n".join(plabel)
		plt.annotate(s=plabel, xy=(.95,.03), xycoords='axes fraction', ha='right', va='bottom', size=8)

	## manual param label
	if False:
		plabel = [r"$\tau_{ev}=%s$"%(Tevap), r"$\tau_{acc}=%s$"%(Tacc)]
		plabel += [r"", r"$M=10^{-3}$", r"$l_{ev}=l=10^{-4}$"]
		plabel = "\n".join(plabel)
		plt.annotate(s=plabel, xy=(.95,.03), xycoords='axes fraction', ha='right', va='bottom', size=8)

	## r=infty label
	if True:
		plt.annotate(s=r"$r=0$",      xy=(.95, 0.), ha='center', va='center', size=7, rotation=90)
		plt.annotate(s=r"$r=\infty$", xy=(1.5, 0.), ha='center', va='center', size=7, rotation=-90)

	## mass plot
	mpgo = True
	if mpgo:
		massplot(chainparams, params)

	## save
	if True:
		xh.evap.evapsave(path=path, params=params, chainparams=chainparams, seed=seed, sfp=sfp, temp_only=temp_only, massplot=mpgo)


	## show
	if False:
		plt.show()


	## return
	return reglist, chainparams




if __name__=="__main__":
	demo()






"""
This module provides test routines for verifying that metfunc object properties have been
properly calculated.
"""

import numpy as np
import matplotlib.pyplot as plt




def func_test(func, saveplots=False):
	"""
	Runs a series of tests to verify that metfunc object properties have been properly calculated.
	"""
	print('\nRUN FUNC TEST')
	print('\n' + '\n'.join(['%s: %s'%(key,func.info[key]) for key in func.info.keys()]) +'\n')
	f_test(func, saveplots)
	F_test_0(func, saveplots)
	F_test_1(func, saveplots)
	F_test_2(func, saveplots)
	#print('\n' + '\n'.join(['%s: %s'%(key,func.info[key]) for key in func.info.keys()]) +'\n')
	print('END FUNC TEST\n')





def nice_plot_background(func):
	"""
	Fill the intervals I_j, with blue as trapped and green as untrapped.
	Separate intervals with vertical grey lines.
	Plot a horizontal black dotted line at y=0.
	"""
	## styles
	fillparams = [dict(facecolor=r'b', alpha=.15, edgecolor='none', zorder=1),
				  dict(facecolor=r'g', alpha=.2 , edgecolor='none', zorder=1)]
	vert_sty = dict(ls='-', c='0.8', lw=1, zorder=5)
	zero_sty = dict(c='k',lw=.5,ls=':',zorder=3)
	##
	rj = func.rj
	ri = rj[1:-1]
	## fill intervals
	for i in range(len(rj)-1):
		plt.fill_betweenx([-100,100],rj[i],rj[i+1], **fillparams[(i+func.first_trapped_j) % 2])
	## plot interval separators
	for i in range(len(ri)):
		plt.plot([ri[i],ri[i]],[-100.,100.], **vert_sty)
	## plot zero line
	plt.plot([0,100],[0,0], **zero_sty)






def f_test(func, saveout=False):
	"""
	Plot metfunc.f(r) along with metfunc.rj and metfunc.kj to check that zeroes and slopes are accurate.
	"""
	## styles
	f_sty = dict(c='k', ls='-', lw=2, zorder=50, label=r'f(r)')
	tang_sty = dict(ls='-', c='0.8', lw=1, zorder=5, label='tangent at rj')
	## make figure and background
	plt.figure()
	nice_plot_background(func)
	## plot metric function
	rr = func.r_ref
	ff = func.f(rr)
	plt.plot(rr,ff,**f_sty)
	## plot tangents
	ri = func.rj[1:-1]
	ki = func.kj[1:-1]
	for i in range(len(ri)):
		plt.plot(rr, ki[i]*(rr-ri[i]), **tang_sty)
		if i==0:
			tang_sty.pop('label')
	## format
	plt.title('f_test: check zeroes and slopes of f(r) at r_i')
	plt.xlabel('r')
	plt.ylabel('f(r)')
	plt.xlim(0, np.max(rr))
	plt.ylim(-10,10 )
	plt.legend(loc='lower right', fontsize=10)
	if saveout==True:
		plt.savefig('f_test.pdf')
	plt.show()




def F_test_0(func, saveout=False):
	"""
	Plot metfunc.F(r) and metfunc.Finv(j,rstar) on top of metfunc.f(r) and metfunc.rj.
	Check that Finv is truly the inverse of F.
	"""
	## styles
	F_sty = dict(c='b', ls='-', lw=2, zorder=50, label='F(r)')
	Finv_sty = dict(ls='none', marker='x', c='r', lw=1, zorder=40, label='Finv(j,rstar)')
	f_sty = dict(c='0.8', ls='-', lw=1, zorder=10, label='f(r)')
	## make figure and background
	plt.figure()
	nice_plot_background(func)
	## plot
	rr = func.r_ref
	rrstar = func.F(rr)
	jj = func.winterval(rr)
	plt.plot(rr,rrstar,**F_sty)
	plt.plot(func.Finv(jj,rrstar),rrstar,**Finv_sty)
	plt.plot(rr,func.f(rr),**f_sty)
	## format
	plt.title('F_test_0: test inverse relation of F and Finv')
	plt.xlabel('r')
	plt.ylabel('')
	plt.xlim(0, np.max(rr))
	plt.ylim(-10,10 )
	plt.legend(loc='lower right', fontsize=10)
	if saveout==True:
		plt.savefig('F_test_0.pdf')
	plt.show()




def F_test_1(func, saveout=False):
	"""
	Plot the reciprocal numerical derivative dF/dr on top of f(r).
	Check that 1./(dF/dr) = f(r) so that tortoise function is an integral of f(r)^{-1}.
	"""
	## styles
	F_sty = dict(c='0.5', ls='-', lw=1.5, zorder=10, label='F(r)')
	dFdr_sty = dict(c='r', ls='none', marker='x', zorder=20, label='numerical 1./(dF/dr)')
	f_sty = dict(c='b', ls='-', lw=1.5, zorder=40, label='f(r)')
	## make figure and background
	plt.figure()
	nice_plot_background(func)
	## plot f(r) and F(r)
	rr = func.r_ref
	rrstar = func.F(rr)
	plt.plot(rr,rrstar,**F_sty)
	plt.plot(rr,func.f(rr),**f_sty)
	## get derivative dF/dr by right difference
	dF = np.roll(rrstar,-1) - rrstar
	dF[-1] = np.nan
	dr = np.roll(rr,-1) - rr
	dFdr = dF / dr
	## plot reciprocal derivative 1./dFdr vs r omitting final point
	plt.plot(rr, dFdr**(-1), **dFdr_sty)
	## format
	plt.title('F_test_1: test dF/dr = 1/f(r)')
	plt.xlabel('r')
	plt.ylabel('')
	plt.xlim(0, np.max(rr))
	plt.ylim(-10,10 )
	plt.legend(loc='lower right', fontsize=10)
	if saveout==True:
		plt.savefig('F_test_1.pdf')
	plt.show()


def F_test_2(func, saveout=False):
	"""
	Plot the continuous version of F(r), with logarithmic infinities subtracted out.
	Plot the exponential function appearing in the Penrose coordinate metric.
	Check continuity at ri.
	"""
	## styles
	F_sty = dict(c='0.7', ls='-', lw=1.5, zorder=10, label=r'$F(r)$')
	Fcont_sty = dict(c='k', ls='-', lw=4, zorder=50, label=r'$F_{cont}(r) = F(r) - \sum_i \, k_i^{-1} \ln(r-r_i)$')
	Fexp_sty = dict(ls='-', lw=2.5, zorder=40)
	Fexp_cols = ['m','c','g','r','b']
	## make figure and background
	plt.figure()
	nice_plot_background(func)
	## plot F(r)
	rr = func.r_ref
	rrstar = func.F(rr)
	jj = func.winterval(rr)
	plt.plot(rr,rrstar,**F_sty)
	## get ri, ki
	ri = func.rj[1:-1]
	ki = func.kj[1:-1]
	## get and plot Fcont = F with logarithmic infinities subtracted out
	Fcont = rrstar
	for i in range(len(ri)):
		Fcont = Fcont - ki[i]**(-1) * np.log( np.abs( rr-ri[i] ) )
	plt.plot(rr, Fcont, **Fcont_sty)
	## get and plot exponential analytic functions
	for i in range(len(ri)):
		Fexp = np.abs( func.f(rr) ) * np.exp( - ki[i] * rrstar )
		Fexp_sty.update(c=Fexp_cols[i%len(ri)], label=r'$F_{exp}(r) = |f(r)| e^{-k_i F(r)}, \quad (i=%d)$'%(i+1) )
		plt.plot(rr, Fexp, **Fexp_sty)
		Fexp_sty['zorder'] += -1
	## format
	plt.title('F_test_2: check continuity properties of F(r)')
	plt.xlabel('r')
	plt.ylabel('')
	plt.xlim(0, np.max(rr))
	plt.ylim(-10,10 )
	plt.legend(loc='lower right', fontsize=11)
	if saveout==True:
		plt.savefig('F_test_2.pdf')
	plt.show()





"""
This module contains functious for generating useful curves which can be added to a diagram.

DEFAULT ZORDER:
100 = FILL
200 = UV GRID
400 = BOUNDARY
600 = RLINES


"""

import numpy as np

from xhorizon.diagram_tools.curve_class import curve
from xhorizon.diagram_tools.block_masks import rstar_limits





#############################  rlines ###############################

def rlines(rvals, sty={}, inf=50., t0=0., npoints=1000):
	"""
	Produce lines of constant radius for r in rvals. Return list of curves.
	"""
	## default style
	style = dict(zorder=600)
	## go
	crvlist = []
	for rval in rvals:
		## create new curve
		crv = curve()
		## define t and r arrays
		t = t0 + np.linspace(-inf, inf, 2*npoints+1)
		r = rval + 0.*t
		## add coords to curve
		crv.tr = np.array([t,r])
		## set style
		style.update(sty)
		crv.sty.update(style)
		## add crv to output list
		crvlist += [crv]
	## return
	return crvlist


def default_rlines(func, sty={}, inf=50., npoints=1000):
	## get radius length scales of the problem
	rscales = (func.rj[1:-1])
	## no length cale
	if len(rscales) == 0:
		dr = np.array([0.5])
		rr = np.array([0.,20.])
	## one length scale
	if len(rscales) == 1:
		dr = rscales / 10.
		rr = np.array([0.5*dr[0], 20.*rscales[0]])
	## multiple length scales
	if len(rscales) >= 2:
		dr = rscales / 10.
		rr = np.zeros(len(rscales)+1)
		rr[1:] = 5.*rscales[:]
		rr[-1] = 20.*rscales[-1]
		rr[0] = 0.5*rscales[0]
	## create curves
	cols = ['b', 'g']
	crvlist = []
	for i in range(len(rr)-1):
		imax = len(rr)-2
		col = cols[(imax-i)%len(cols)]
		rvals = np.arange(rr[i], rr[i+1], dr[i])
		style = dict(c=col, zorder=599-i)
		style.update(sty)
		crvlist += rlines(rvals, sty=style, inf=inf, npoints=npoints)
	## return
	return crvlist

#####################################################################


#############################  rstarlines ###############################

def rstarlines(vals, c=None, sty={}, inf=50., npoints=1000):
	"""
	Produce lines of constant rstar for rstar in vals. Return list of curves.
	"""
	crvlist = []
	for val in vals:
		## create new curve
		crv = curve()
		## define t array
		t = np.linspace(-inf, inf, 2*npoints+1)
		## define u and v arrays
		c = float(c)
		u = t - val + c
		v = t + val - c
		## add coords to curve
		crv.uv = np.array([u,v])
		## update curve style from default
		style = dict(zorder=600)
		style.update(sty)
		crv.sty.update(style)
		## add crv to output list
		crvlist += [crv]
	## return
	return crvlist



def rstarlines_special_1(vals, c=None,  s0=None, ki=None, A=1e4, sty={}, inf=100., npoints1=1001, npoints2=1001, npoints3=1001):
	"""
	Produce lines of constant rstar for rstar in vals. Return list of curves.
	These are specially supplemented to give many data points immediately after s0 for zoom views.
	"""
	crvlist = []
	for val in vals:
		## initialize params
		c = float(c)
		## create new curve
		crv = curve()
		## initialize u and v lists
		ulist = []
		vlist = []
		## small values
		ss = np.linspace(-1.02*s0,1.02*s0,npoints1)
		for x in [2.*ss]:
			## u values
			ulist += [x]
			vlist += [x + 2.*(val-c)]
			## v values
			vlist += [-x]
			ulist += [-x - 2.*(val-c)]
		## large values
		ss = np.linspace(s0,inf,npoints3)
		for x in [2.*ss, -2.*ss]:
			## u values
			ulist += [x]
			vlist += [x + 2.*(val-c)]
			## v values
			vlist += [-x]
			ulist += [-x - 2.*(val-c)]
		## prep intermediate values
		e = np.sign(ki)
		ds = ki**(-1) * np.log(1.+e*ki*A)
		np2 = [ int(np.abs(e[i])) * npoints2 for i in range(len(ki)) ]
		ss = [ np.linspace(e[i]*s0, e[i]*s0+ds[i], np2[i]) for i in range(len(ki)) ]
		## intermediate values
		for x in [2.*ss[0], 2.*ss[1]]:
			## u values
			ulist += [x]
			vlist += [x + 2.*(val-c)]
			## v values
			vlist += [-x]
			ulist += [-x - 2.*(val-c)]
		## concatenate to form output
		u = np.concatenate(ulist)
		v = np.concatenate(vlist)
		## sort
		mask = np.argsort(u)
		u = u[mask]
		v = v[mask]
		## add coords to curve
		crv.uv = np.array([u,v])
		## update curve style from default
		style = dict(zorder=600)
		style.update(sty)
		crv.sty.update(style)
		## add crv to output list
		crvlist += [crv]
	## return
	return crvlist


def rstarlines_special_2(vals, uvbounds, c=None, sty={}, inf=100., npoints=1000, eps=1e-12):
	"""
	Produce lines of constant rstar for rstar in vals. Return list of curves.
	   (v-u)/2 = rstar - c
	"""
	## initialize
	crvlist = []
	uvb = uvbounds.copy()
	cc = 1.*float(c)
	## go
	for val in vals:
		## create new curve
		crv = curve()
		## adjust uvbounds
		for key in ['vmin', 'umin']:
			if not np.isfinite(uvb[key]):
				uvb[key] = -1.*inf
		for key in ['vmax', 'umax']:
			if not np.isfinite(uvb[key]):
				uvb[key] =  1.*inf
		## readable params
		umin, umax, vmin, vmax = uvb['umin'] , uvb['umax'], uvb['vmin'], uvb['vmax']
		## push inward just a bit to avoid mask
		umin, umax = umin + 1.*eps, umax - 1.*eps, 
		vmin, vmax = vmin + 1.*eps, vmax - 1.*eps
		## curve a
		ua = np.linspace(umin, umax, npoints)
		va = ua + 2.*(val - 1.*cc)
		## curve b
		vb = np.linspace(vmin, vmax, npoints)
		ub = vb - 2.*(val-1.*cc)
		## combine
		uu = np.concatenate([ua,ub])
		vv = np.concatenate([va,vb])
		## sort
		idx = np.argsort(uu)
		uu = uu[idx]
		vv = vv[idx]
		## add coords to curve
		crv.uv = np.array([uu,vv])
		## update curve style from default
		style = dict(zorder=600)
		style.update(sty)
		crv.sty.update(style)
		## add crv to output list
		crvlist += [crv]
	## return
	return crvlist


#####################################################################

############################# uvlines  ###############################

def uvlines(uvvals, uv='uv', uvbounds=dict(), sty={}, c=0., eps=1e-24, inf=100., npoints=1001):
	"""
	Produce lines of constant radius for r in rvals. Return list of curves.
	"""
	crvlist = []
	for val in uvvals:
		## create new curve
		ucrv = curve()
		vcrv = curve()
		## adjust uvbounds
		uvb = dict(vmin=np.nan, vmax=np.nan, umin=np.nan, umax=np.nan)
		uvb.update(uvbounds)
		for key in ['vmin', 'umin']:
			if not np.isfinite(uvb[key]):
				uvb[key] = -1.*inf
		for key in ['vmax', 'umax']:
			if not np.isfinite(uvb[key]):
				uvb[key] =  1.*inf
		## readable params
		umin, umax, vmin, vmax = uvb['umin'] , uvb['umax'], uvb['vmin'], uvb['vmax']
		## push inward just a bit to avoid mask
		umin, umax = umin + 1.*eps, umax - 1.*eps, 
		vmin, vmax = vmin + 1.*eps, vmax - 1.*eps
		## define linearly spaced array
		sa1 = np.linspace(umin, umax, npoints)
		sa2 = np.linspace(vmin, vmax, npoints)
		sb = val - 2.*c + eps*np.array([-1.,0.,1.])
		sc = val + 2.*c + eps*np.array([-1.,0.,1.])
		s = np.sort(np.concatenate([sa1,sa2,sb,sc]))
		## ucrv is line of const, vcrv is line of const v
		ucrv.uv = np.array([val+0.*s, 1.*s])
		vcrv.uv = np.array([1.*s, val+0.*s])
		## update curve style from default
		style = dict(zorder=200)
		style.update(sty)
		ucrv.sty.update(style)
		vcrv.sty.update(style)
		## add crvs to output list
		if 'u' in uv:
			crvlist += [ucrv]
		if 'v' in uv:
			crvlist += [vcrv]
	## return
	return crvlist

#####################################################################

############################# boundary  ###############################


def block_boundary(blk, sty={}, inf=100., npoints=1000):
	"""
	"""
	## change default style
	style = dict(c='k', ls='-', lw=1.5, zorder=400)
	style.update(sty)
	## get min and max rstar values
	j = blk.j
	lims = rstar_limits(blk.master.metfunc)
	lims = lims[j:j+2]
	## adjust to avoid accidental masking
	idx = np.argsort(lims)
	lims[idx] = lims[idx] + 2. * np.array([1.,-1.]) * blk.master.metfunc.Fparams['eps']
	## make list of boundary curves
	crvlist = []
	## make curves
	for rstar in lims:
		if rstar==np.inf:
			rstar = inf
		if rstar==-np.inf:
			rstar = -inf
		crvlist += rstarlines([rstar], c=blk.master.rparams['c'], sty=style, inf=2.*inf, npoints=npoints)
	## return
	return crvlist




def block_boundary_2(blk, sty={}, inf=100., npoints=2001):
	"""
	"""
	## change default style
	style = dict(c='k', ls='-', lw=1.5, zorder=400)
	style.update(sty)
	## get min and max rstar values
	j = blk.j
	lims = rstar_limits(blk.master.metfunc)
	lims = lims[j:j+2]
	## adjust to avoid accidental masking
	idx = np.argsort(lims)
	lims[idx] = lims[idx] + 2. * np.array([1.,-1.]) * blk.master.metfunc.Fparams['eps']
	## make list of boundary curves
	crvlist = []
	## make curves
	for rstar in lims:
		if rstar==np.inf:
			rstar = inf
		if rstar==-np.inf:
			rstar = -inf
		crvlist += rstarlines_special_2([rstar], blk.uvbounds, c=blk.master.rparams['c'], sty=style, inf=inf, npoints=npoints)
	## return
	return crvlist



#####################################################################

############################# other  ###############################

def colnorm1(svals,smin,smax):
	"""
	Color bar normalizer for input into colormap.
	Input an array of svals.
	Return an array of the absolute values normalized to (0,1) and truncated at smin, smax.
	"""
	## initialize output array to abs val of input
	ss = 1. * np.abs(svals)
	## cutoff at min
	ss[ss<smin] = smin + 0.*ss[ss<smin]
	## cutoff at max	
	ss[ss>smax] = smax + 0.*ss[ss>smax]
	## normalize
	ss = (ss - smin) / (smax - smin)
	## return
	return ss


#####################################################################



























"""
This module defines a class used for evaluating coordinate transformations at null shell junctions.
"""


import numpy as np
from xhorizon.shell_junction import interpolators as interp
from xhorizon.shell_junction.helpers import *




class active_slice:

	"""
	Class for handling shell and corner slicing of SSS regions. Given the region and the slice parameters,
	reference arrays are created for all desired transformations, which are then defined by interpolation
	and extrapolation of the reference arrays.

	Unlike passive_slice, this will actively obtain new functions U(udl),V(vdl) based on the inputs U0,V0, rather
	than just reading the existing ones.

	Designed to simultaneously incorporate both shell and corner junctions. Input determines which behavior will
	take effect. Offers protection from bad inputs in most cases. The exception is that the correct ublocks and 
	vblocks have to be given or everything will come out whacky.

	Slice Location:
		The location of the slice is determined by the values r0,u0,v0. These describe a point with coordinates
		(u0,v0) in some block, and with radius r0. The values are forced to be self-consistent by the algorithm.
		In corner junction mode, this is the corner point. Otherwise, it just describes a point on the shell.
		In shell mode, r0 and (u0 xor v0) can be nan, in which case the parameters just specify the shell.
		To fully specify the point we need to know what block it's in. So correct ublock and vblock args are needed.

	Slice Location Input Requirements:
		u0==finite xor v0==finite                   # avoids overspecifying a nonphysical point
		(optional in shell mode) r0==finite         # if finite, used to determine the other one of u0,v0

	Corner Junction Mode Input Requirements:
		r0==finite                  # corner junction point radius
		u0==finite xor v0==finite   # corner junction coordinate value
		U0 not None                 # function U0(r) at v=v0
		V0 not None                 # function V0(r) at u=u0
		ublocks==[correct]          # the correct list of block indices must be provided
		vblocks==[correct]          # the correct list of block indices must be provided

	Shell Junction Mode Input Requirements:
		For example, suppose we want a shell junction at v=v0.
			(v0==finite) xor (r0==finite and u0==finite)  # either of these uniquely determines the value v=v0 for the shell
			U0 not None                                   # function U0(r) at v=v0 shell
			vblocks==[correct]                            # the correct list of block indices must be provided
		For shell at u=u0, just switch u's and v's.

	Inputs:
		reg = The region being sliced.
		ublocks = List of indices (corresponding to reg.blocks) for the blocks containing the u=u0 slice.
		vblocks = List of indices (corresponding to reg.blocks) for the blocks containing the v=v0 slice.
		r0 = Radius of the junction corner point. Or, radius of the specified point on the shell in shell mode. See above.
		u0 = Slice location coordinate. See above.
		v0 = Slice location coordinate. See above.
		U0 = Function U0(r) at v=v0, or None. When None, U(udl) = udl.
		V0 = Function V0(r) at u=u0, or None. When None, V(vdl) = vdl.
		mu = Extrapolation parameter. When mu=0, extrapolation is just translation. When mu->inf, extrapolation is smooth linear. 
				(See interpolators module).

	Methods:
		Provides various coordinate transformation methods, which can be used by a region.

	Attributes:
		Input parameters as well as various reference arrays.

	"""

	def __init__(self, reg, ublocks=[], vblocks=[], r0=np.nan, u0=np.nan, v0=np.nan, U0=None, V0=None, mu=0., r_refs=[]):

		## process and store input values
		self.reg = reg
		self.ublocks, self.vblocks = bcon(reg,ublocks), bcon(reg,vblocks)
		self.r0, self.u0, self.v0 = set_ruv0(reg, r0=r0, u0=u0, v0=v0)
		self.U0, self.V0 = U0, V0
		self.mu = 1.*float(mu)

		## get ref arrays
		self.r = get_r_ref(self.reg, r_refs, self.r0)
		self.r, self.uvdl_u0, self.uvdl_v0 = uvdl_of_r_at_uv0(self.r, self.reg, ublocks=self.ublocks, vblocks=self.vblocks, u0=self.u0, v0=self.v0)
		self.U_v0 = U_of_udl_at_v0(self.r, self.uvdl_v0[0], self.U0)
		self.V_u0 = V_of_vdl_at_u0(self.r, self.uvdl_u0[1], self.V0)

		## use interpolated functions to get UV of ref arrays
		if np.isfinite(r0) and len(ublocks)>0 and len(vblocks)>0:
			self.UV_u0 = self.UV_of_uvdl(self.uvdl_u0)
			self.UV_v0 = self.UV_of_uvdl(self.uvdl_v0)

	"""
	Coordinate transformation methods.
	"""

	def U_of_udl_at_v0(self, udl):
		"""
		Evaluate the function U(udl) = U(r(udl,vdl0)) by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		"""
		return interp.interp_with_smooth_extrap(udl, self.uvdl_v0[0], self.U_v0, mu=self.mu)

	def V_of_vdl_at_u0(self, vdl):
		"""
		Evaluate the function V(vdl) = V(r(udl0,vdl)) by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		"""
		return interp.interp_with_smooth_extrap(vdl, self.uvdl_u0[1], self.V_u0, mu=self.mu)

	def UV_of_uvdl(self, uvdl):
		"""
		Combine U(udl) and V(vdl) into UV(uvdl).
		"""
		U_temp = self.U_of_udl_at_v0(uvdl[0])
		V_temp = self.V_of_vdl_at_u0(uvdl[1])
		UV_temp = np.array([U_temp,V_temp])
		return UV_temp

	def U_of_r_at_v0(self, r):
		"""
		Evaluate the function U(r) at v0 by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		Setting mu to nan because this should not be extrapolated.
		"""
		return interp.interp_with_smooth_extrap(r, self.r, self.U_v0, mu=np.nan)

	def V_of_r_at_u0(self, r):
		"""
		Evaluate the function V(r) at u0 by interpolating from stored reference values.
		Smooth interpolation and extrapolation handled by interpolators module.
		Setting mu to nan because this should not be extrapolated.
		"""
		return interp.interp_with_smooth_extrap(r, self.r, self.V_u0, mu=np.nan)










"""
Hello! I am the docstring heading the top level __init__.py in xhorizon.
"""

######### FROM SKELETON ###########

from importlib.metadata import version
__version__ = version("xhorizon")

###################################

## external
import numpy as np
import matplotlib.pyplot as plt
plt.style.use("classic")

## main subpackages
from xhorizon import tortoise
from xhorizon import diagram_tools
from xhorizon import shell_junction

## convenience names
from xhorizon.tortoise import metfunc_factory as mf
from xhorizon.diagram_tools import region_factory as reg
from xhorizon.diagram_tools import curvemakers as cm

## alternate sub names
from xhorizon import shell_junction as junc

## more convenience names
from xhorizon.diagram_tools.plotstyle import newfig, pubrc
from xhorizon.diagram_tools.curve_class import curve
from xhorizon.diagram_tools.diagram_classes import diagram, region, block
from xhorizon.tortoise.metfunc_tests import func_test

## more convenience names
from xhorizon.diagram_tools import block_masks as mask
from xhorizon.diagram_tools import block_fill as fill
from xhorizon.shell_junction import reg_corner_masks as cornermask


## evap subpackage needs to be loaded after the above due to circular import issue
from xhorizon import evap





"""
This module specifies standard transformations between
the double null block coordinates $(u,v)$ and region
coordinates $(\tilde{u},\tilde{v})$ (the standard
Penrose coordinates on an SSS region), along with the
associated helper functions.
"""

"""
The coordinate hierarchy is as follows:

"Schwarzschild Coordinates"
tr = $ (t,r) $
Schwarzschild coordinates covering a single block with metric
$ ds^2 = -f(r) dt^2 + f(r)^{-1} dr^2 + r^2 d\Omega^2 $.

"Double Null Block Coordinates"
uv = $ (u,v) $
Double-null block coordinates covering a single block with metric
$ ds^2 = - du dv + r^2 d\Omega^2 $.

"Region Coordinates"
uvdl = $ (\tilde{u},\tilde{v}) $
Standard Penrose coordinates for an SSS spacetime.

"Diagram Coordinates"
UV = $ (U,V) $
Coordinate system to be directly plotted in Penrose diagram.
Obtained from region coords by transformations U(udl) and V(vdl).
"""


import numpy as np


## block to block transformations

def uv_of_tr(tr, F, c):
	t, r = tr
	u = t - F(r) + c
	v = t + F(r) - c
	uv = np.array([u,v])
	return uv

def tr_of_uv(uv, Finv, c):
	u, v = uv
	t = 0.5*(v+u)
	r = Finv( 0.5*(v-u) + c)
	tr = np.array([t,r])
	return tr

def r_of_uv(uv, Finv, c):
	r = (tr_of_uv(uv, Finv, c))[1]
	return r



## block to region transformations

def uvdl_of_uv(uv, cdlu, cdlv, epsu, epsv, kpm, s0):
	h = lambda s: hks(s, kpm, s0)
	u, v = uv
	udl = np.pi**(-1) * np.arctan(  epsu * h( u/2.) ) + cdlu
	vdl = np.pi**(-1) * np.arctan( -epsv * h(-v/2.) ) + cdlv
	uvdl = np.array([udl,vdl])
	return uvdl

def uv_of_uvdl(uvdl, cdlu, cdlv, epsu, epsv, kpm, s0):
	hinv = lambda s: hksinv(s, kpm, s0)
	udl, vdl = uvdl
	u =  2. * hinv(  epsu**(-1) * np.tan( np.pi * (udl - cdlu) ) )
	v = -2. * hinv( -epsv**(-1) * np.tan( np.pi * (vdl - cdlv) ) )
	uv = np.array([u,v])
	return uv











## supplementary functions

def hks(s, kpm, s0):
	## get k_+ and k_- values
	km, kp = np.min(kpm), np.max(kpm)
	## initialize array w nan
	h = np.nan * s
	## implement piecewise function
	mask = s < -s0
	h[mask] = -s0 + H(s[mask]+s0, km)
	mask = np.abs(s) <= s0
	h[mask] = s[mask]
	mask = s > s0
	h[mask] =  s0 + H(s[mask]-s0, kp)
	## return
	return h

def H(s,k):
	## initialize array w nan
	H = np.nan * s
	## implement piecewise function
	if k!=0.:
		H = k**(-1) * ( np.exp(k*s) - 1. )
	if k==0.:
		H = s
	## return
	return H


def Hinv(s,k):
	## initialize array w nan
	Hinv = np.nan * s
	## implement piecewise function
	if k!=0.:
		Hinv = k**(-1) * np.log( k*s + 1 )
	if k==0.:
		Hinv = s
	## return
	return Hinv


def hksinv(s, kpm, s0):
	## get k_+ and k_- values
	km, kp = np.min(kpm), np.max(kpm)
	## initialize array w nan
	hinv = np.nan * s
	## implement piecewise function
	mask = s < -s0
	hinv[mask] = -s0 + Hinv(s[mask]+s0, km)
	mask = np.abs(s) <= s0
	hinv[mask] = s[mask]
	mask = s > s0
	hinv[mask] =  s0 + Hinv(s[mask]-s0, kp)
	## return
	return hinv













"""
Tests. Tests run if __name__=="__main__".
"""

import matplotlib.pyplot as plt


def test1():
	"""
	Test that uv_of_tr and tr_of_uv are stable numerical inverses.
	"""
	print( "\nTEST 1")
	## input ##
	npts = 100
	t = 10. * ( 2.*np.random.rand(npts) - 1. )
	r = 10. * np.random.rand(npts)
	c = 10. * ( 2.*np.random.random() - 1. )
	F = lambda rr: np.arctan(rr)
	Finv = lambda rrstar: np.tan(rrstar)
	##
	## test block to block transformations
	tr = np.array([t,r])
	uv = uv_of_tr(tr, F, c)
	uvnew = uv_of_tr(tr, F, c)
	trnew = tr_of_uv(uv, Finv, c)
	for i in range(4):
		uvnew = uv_of_tr(trnew, F, c)
		trnew = tr_of_uv(uvnew, Finv, c)
		rnew = r_of_uv(uvnew, Finv, c)
	## results of test
	if False:
		print( "uv = ", uv)
		print( "tr = ", tr)
		print( "r = ", r)
	if True:
		print( "max diff uv = ", np.max(np.abs(uvnew-uv)))
		print( "max diff tr = ", np.max(np.abs(trnew-tr)))
		print( "max diff r = ", np.max(np.abs(rnew-r)))
	## end
	print( "END TEST 1\n")


def test2():
	"""
	Test that H and Hinv are stable numerical inverses.
	"""
	print( "\nTEST 2")
	## input ##
	s = np.linspace(-20.,20.,3001)
	k = 1.
	##
	## test H
	x = s
	y = H(s, k)
	xnew = Hinv(y, k)
	ynew = H(xnew, k)
	for i in range(4):
		xnew = Hinv(ynew, k)
		ynew = H(xnew, k)
	if False:
		print( "x = ", x)
		print( "y = ", y)
	if True:
		print( "max diff x = ", np.max(np.abs(xnew-x)))
		print( "max diff y = ", np.max(np.abs(ynew-y)))
	## end
	print( "END TEST 2\n")


def test3():
	"""
	Test that h and hinv are stable numerical inverses.
	"""
	print( "\nTEST 3")
	## input ##
	s = np.linspace(-20.,20.,2001)
	kpm = np.array([-1.,2.])
	s0 = 5.
	## setup h
	h = lambda s: hks(s, kpm, s0)
	hinv = lambda s: hksinv(s, kpm, s0)
	## test h
	x = s
	y = h(s)
	xnew = hinv(y)
	ynew = h(xnew)
	for i in range(4):
		xnew = hinv(ynew)
		ynew = h(xnew)
	if False:
		print( "x = ", x)
		print( "y = ", y)
	if True:
		print( "max diff x = ", np.max(np.abs(xnew-x)))
		print( "max diff y = ", np.max(np.abs(ynew-y)))
	## end
	print( "END TEST 3\n")


def test4():
	"""
	Plot h and hinv.
	"""
	print( "\nTEST 4")
	## input ##
	s = np.linspace(-20.,20.,2001)
	kpm = np.array([-0.6,0.2])
	kpm = np.sort(kpm)
	s0 = 5.
	## setup h
	h = lambda s: hks(s, kpm, s0)
	hinv = lambda s: hksinv(s, kpm, s0)
	## test plot h
	plt.figure()
	plt.plot(s,s,'0.8')
	plt.plot(s,h(s),'k-', label='hks(s,kpm,s0)')
	plt.plot(s,hinv(s),'k--', label='hksinv(s,kpm,s0)')
	sq = 2.*s0
	plt.xlim(-sq,sq)
	plt.ylim(-sq,sq)
	plt.xticks([-s0,0,s0])
	plt.yticks([-s0,0,s0])
	plt.grid()
	plt.gca().set_aspect('equal')
	plt.xlabel('s')
	plt.title('TEST 4 (k-=%s, k+=%s, s0=%s)'%(kpm[0],kpm[1],s0))
	plt.legend(loc='lower right', fontsize=10)
	plt.show()
	## end
	print( "END TEST 4\n")


def test5():
	"""
	Plot H and Hinv.
	"""
	print( "\nTEST 5")
	## input ##
	s = np.linspace(-20.,20.,10001)
	k = -10.
	## test plot H
	plt.figure()
	plt.plot(s,s,'0.8')
	plt.plot(s, H(s,k),'k-', label='H(s,k)')
	## test plot Hinv
	s = np.linspace(-.999999/k, 10.*k, 10001)
	plt.plot(s, Hinv(s,k), 'k--', label='Hinv(s,k)')
	## format
	sq = 2.
	plt.xlim(-sq,sq)
	plt.ylim(-sq,sq)
	plt.grid()
	plt.gca().set_aspect('equal')
	plt.xlabel('s')
	plt.title('TEST 5 (k=%r)'%(k))
	plt.legend(loc='lower right', fontsize=10)
	plt.show()
	## end
	print( "END TEST 5\n")

## run test routines if this module is main

if __name__=='__main__':
	test1()
	test2()
	test3()
	test4()
	test5()





"""

"""

import numpy as np

from xhorizon.diagram_tools.curve_class import curve


"""
Specific masks.
Each of these subs returns
"""

def uv_range(blk, crvlist):
	"""
	"""
	## get uvbounds from block
	uv=dict(umin=-np.inf, umax=np.inf, vmin=-np.inf, vmax=np.inf)
	uv.update(blk.uvbounds)
	umin, umax = uv['umin'], uv['umax']
	vmin, vmax = uv['vmin'], uv['vmax']
	## go
	out = []
	for crv in crvlist:
		u,v = crv.uv
		mask = np.logical_and.reduce( [umin<=u, u<=umax, vmin<=v, v<=vmax] )
		out += mask_split_curve(crv, mask)
	return out

def rstar_minmax(blk, crvlist):
	"""
	Allow only valid values of rstar in the given block.
	"""
	## get min and max rstar values
	j = blk.j
	lims = rstar_limits(blk.master.metfunc)
	left, right = lims[j], lims[j+1]
	rstar_min, rstar_max = np.sort(np.array([left, right]))
	## proceed to mask curves
	out = []
	for crv in crvlist:
		u, v = crv.uv
		rstar = 0.5*(v-u) + blk.master.rparams['c']
		mask = np.logical_and.reduce( [rstar_min<=rstar, rstar<=rstar_max] )
		out += mask_split_curve(crv, mask)
	## return
	return out


def rstar_limits(func):
	"""
	Find limiting values of rstar at each rj given the metfunc.
	"""
	## get N
	N = len(func.rj) - 2
	## first N-1 blocks always follow same pattern
	rstar_limits = - np.inf * np.sign(func.kj)
	rstar_limits[0] = 0.
	## final block is trickier
	inf = 1e3 * func.rj[-1]
	finf = func.f(inf)
	## default to +- inf
	rstar_limits[N+1] = finf * np.inf
	## if |f(r)|>r as r to inf then |1./f(r)| < 1./r as r to inf, so F(r) converges
	## if F(r) converges use last computed value
	if np.abs(finf) > inf:
		rstar_limits[N+1] = (func.rstar_ref[np.argsort(func.r_ref)])[-1]
	## return
	return rstar_limits





"""
General tools.
"""

def mask_split_array(array, mask):
	"""
	Given an array and a mask of proper length, split the array into valid subarrays.
	Return list of valid subarrays. Discard invalid subarrays.
	Apply mask to itself to get length info.
	If array is empty, split into the correct number of empty arrays.
	"""
	## mask to int
	mask = mask.astype(int)
	## prepare output list of arrays
	out = []
	## get split indices for subarrays
	idx = np.nonzero(mask[1:] - mask[:-1])[0] + 1
	idx0 = np.concatenate([np.array([0]),idx])
	Nsubs = len(idx0)
	## if array is filled, split and return valid subs
	if len(array)>0:
		## split input array into subarrays
		subs = np.split(array, idx)
		## choose valid arrays and add to output
		for i in range(Nsubs):
			if mask[idx0[i]]==True:
				out += [subs[i]]
	## if array is empty, return correct number of empty subarrays
	if len(array)==0:
		for i in range(Nsubs):
			if mask[idx0[i]]==True:
				out += [np.array([])]
	## return
	return out

def mask_split_2darray(array, mask):
	"""
	Applies mask_split_array to each of two columns of a 2d array.
	Input is an array of the form xy = np.array([x,y]) and a boolean mask.
	Output is a list of valid 2d subarrays of xy.
	"""
	## prep output list
	out = []
	## get lists for each axis
	out0 = mask_split_array(array[0], mask)
	out1 = mask_split_array(array[1], mask)
	## recombine
	for i in range(len(out0)):
		out += [ np.array([out0[i],out1[i]]) ]
	## return
	return out


def mask_split_curve(crv, mask):
	"""
	"""
	## prep output list
	out = []
	## get unmasked array values
	r, tr, uv, uvdl, UV  = crv.r, crv.tr, crv.uv, crv.uvdl, crv.UV
	## create list of masked array values
	r  = mask_split_array(r, mask)
	tr = mask_split_2darray(tr, mask)
	uv = mask_split_2darray(uv, mask)
	uvdl = mask_split_2darray(uvdl, mask)
	UV = mask_split_2darray(UV, mask)
	## for each valid subcurve create new curve and fill values
	Nsubs = len(r)
	for i in range(Nsubs):
		newcrv = curve()
		newcrv.sty = crv.sty
		newcrv.r, newcrv.tr, newcrv.uv, newcrv.uvdl, newcrv.UV = r[i], tr[i], uv[i], uvdl[i], UV[i]
		out += [newcrv]
	## return
	return out






















"""
Tests.
Run if __name__='__main__'.
"""



def test1():
	"""
	Demonstrate the functionality of mask_split_array(array, mask).
	"""
	## define input arrays and mask
	x = np.linspace(-10,10,1001)
	y = np.sin(x)
	z = np.array([])
	mask = np.logical_or( np.abs(x)>5., y>.5)
	## get 
	xout = mask_split_array(x,mask)
	yout = mask_split_array(y,mask)
	zout = mask_split_array(z,mask)
	## print outputs
	print(xout)
	print(yout)
	print(zout)
	## plot
	plt.plot(x,y,'b-',lw=15, label='unmasked')
	plt.plot(x[mask],y[mask],'y-',lw=10, label='naive numpy mask')
	label='split mask with mask_split_array'
	for i in range(len(xout)):
		plt.plot(xout[i],yout[i],'r-', lw=5, label=label)
		label=None
	## format
	plt.title('test of mask_split_array(array, mask)')
	plt.legend(loc='lower right', fontsize=11)
	plt.grid()
	plt.show()



def test2():
	"""
	Demonstrate the functionality of mask_split_2darray(array, mask).
	Input is an array of the form xy = np.array([x,y]) and a boolean mask.
	Output is a list of valid 2d subarrays of xy.
	"""
	## define input arrays and mask
	x = np.linspace(-10,10,1001)
	y = np.sin(x)
	z = np.array([])
	mask = np.logical_or( np.abs(x)>5., y>.5)
	## define input 2d arrays
	xy = np.array([x,y])
	xz = np.array([x,z])
	## get 
	xyout = mask_split_2darray(xy,mask)
	xzout = mask_split_2darray(xz,mask)
	## print outputs
	print(xyout)
	print(xzout)
	## plot
	plt.plot(xy[0],xy[1],'b-',lw=15, label='unmasked')
	plt.plot(xy[0][mask],xy[1][mask],'y-',lw=10, label='naive numpy mask')
	label='split mask with mask_split_2darray'
	for i in range(len(xyout)):
		plt.plot(xyout[i][0],xyout[i][1],'r-', lw=5, label=label)
		label=None
	## format
	plt.title('test of mask_split_2darray(array, mask)')
	plt.legend(loc='lower right', fontsize=11)
	plt.grid()
	plt.show()



def test3():
	"""
	Demonstrate the functionality of mask_split_curve(crv, mask).
	"""
	## create input curve
	crv = curve()
	## input curve coordinates
	s = np.linspace(-5,3,1001)
	crv.r = s**2
	crv.tr = np.array([s,s**2])
	crv.uv = np.array([s,s])
	crv.uvdl = np.array([s**3,s**2])
	## define mask
	mask = np.logical_or(np.sin(5.*crv.uv[0]**2) > .5, np.abs(crv.uv[1]) > 3.5)
	## run
	out = mask_split_curve(crv, mask)
	# plot
	for s in ['tr','uv','uvdl','UV']:
		x = crv.__dict__[s]
		## plot unmasked
		plt.plot(x[0], x[1], 'b-', lw=15, label='unmasked')
		## plot naive numpy mask application
		if len(x[0])>0:
			plt.plot(x[0][mask], x[1][mask], 'y-', lw=10, label='naive numpy mask')
		## plot valid subcurves
		label = 'split mask with mask_split_curve'
		for i in range(len(out)):
			x = out[i].__dict__[s]
			plt.plot(x[0], x[1], 'r-', lw=5, label=label)
			label = None
		plt.title(s)
		plt.legend(loc='lower left', fontsize=10)
		plt.show()






## run tests if __name__='__main__'.
if __name__=='__main__':
	## import
	import matplotlib.pyplot as plt
	## test
	# test1()
	# test2()
	# test3()





"""
This module defines the classes
	diagram
	region
	block
which are used to construct and plot diagrams.
"""

import numpy as np
import matplotlib.pyplot as plt

from xhorizon.diagram_tools import coord_transf as coord
from xhorizon.diagram_tools.block_masks import rstar_minmax, uv_range
from xhorizon.diagram_tools import block_fill



class diagram:

	"""
	A diagram is fundamentally a collection of curves.
	A diagram object consists of:
		(1) a list of curves in UV coordinates, and 
		(2) a list of sss regions, each of which bring their own curves.
	"""

	def __init__(self):
		self.curves = []
		self.regions = []

	def add_curves_UV(self, crvlist):
		pass

	def dplot(self):
		### "diagram plot"
		for reg in self.regions:
			reg.rplot()




class region:

	"""
	A region object corresponds to a single piece of an sss spacetime with metric function f(r).
	This will often, but not always, be a single Eddington-Finklestein region of the spacetime.
	A region object consists of:
		(1) a metfunc object which determines the metric function f(r) and its associated transformations
		(2) a list of curve objects explicitly belonging to the region
		(3) a list of block objects belonging to the region
		(4) methods providing transformations from region to diagram coordinates
	"""

	def __init__(self, metfunc, rparams={}):
		self.metfunc = metfunc
		self.rparams = dict(c=0., s0=10.)
		self.rparams.update(rparams)
		self.curves = []
		self.blocks = []

	"""
	UV to uvdl transformations.
	"""

	def UV_of_uvdl(self, uvdl):
		udl, vdl = uvdl
		U = self.U_of_udl(udl)
		V = self.V_of_vdl(vdl)
		UV = np.array([U,V])
		return UV

	def uvdl_of_UV(self, UV):
		U, V = UV
		udl = self.udl_of_U(U)
		vdl = self.vdl_of_V(V)
		uvdl = np.array([udl,vdl])
		return uvdl

	"""
	Null direction transformations.
	"""

	## forward

	def U_of_udl(self,udl):
		U = 1. * udl
		return U

	def V_of_vdl(self,vdl):
		V = 1. * vdl
		return V

	## backward

	def udl_of_U(self,U):
		udl = 1. * U
		return udl

	def vdl_of_V(self,V):
		vdl = 1. * V
		return vdl

	"""
	Utilities.
	"""

	def add_block(self, j, bparams):
		"""
		Add a new block to the region, given index j and parameters bparams.
		"""
		b = block(self, j, bparams)
		if b.consistency_check()==True:
			self.blocks += [b]
		return b

	def add_curves_uvdl(self, crvlist):
		pass
	
	"""
	Plotting.
	"""

	def rplot(self):
		### "region plot"
		for b in self.blocks:
			b.bplot()




class block:

	"""
	A block object corresponds to a single sss block.
	It is owned by a region, its "master", which determines its metric function and region params c,s0.

	The block object does most of the heavy lifting for creating a diagram.
	"""

	def __init__(self, master, j, bparams={}):
		## primary attributes
		self.master = master
		self.j = int(j)
		self.bparams = dict(cdlu=0., cdlv=0., epsu=1., epsv=1.)
		self.bparams.update(bparams)
		self.curves = []
		self.masks = []
		self.uvbounds = dict(umin=-np.inf, umax=np.inf, vmin=-np.inf, vmax=np.inf)
		## derived attributes
		self.rj = None
		self.kj = None
		self.kpm = None
		self.sgnf = None
		self.F    = None
		self.Finv = None
		self.s0 = None
		self.c = None
		## coordinate transformations
		self.uv_of_tr = None
		self.tr_of_uv = None
		self.r_of_uv  = None
		self.uvdl_of_uv = None
		self.uv_of_uvdl = None
		## supplementary functions
		self.h = None
		self.hinv = None
		## set up block
		self.masks += [ rstar_minmax, uv_range ]
		self.update()

	"""
	Methods for use by the user.
	"""

	def update(self):
		"""
		Use current block primary attributes to set derived attributes and coordinate transformations.
		Check for consistency.
		"""
		self.update_derived_attributes()
		self.consistency_check()
		self.update_coord_transf()

	def add_curves_uv(self, crvlist):
		"""
		Add a curve to the block, using uv coords to fill all other coordinate fields.
		"""
		self.curves += self.apply_masks(self.update_curves_from_uv(crvlist))

	def refresh_curves_from_uv(self):
		"""
		Recalculate coordinates of all curves based on the current uv coords.
		"""
		self.curves = self.apply_masks(self.update_curves_from_uv(self.curves))

	def add_curves_tr(self, crvlist):
		"""
		Add a curve to the block, using tr coords to fill all other coordinate fields.
		Only curves lying entirely within the block can be added.
		If any points have an invalid radius, whole curve is discarded.
		"""
		self.curves += self.apply_masks(self.update_curves_from_tr(crvlist))

	def update_curves_from_tr(self, crvlist):
		"""
		Update curves to the block, using tr coords to fill all other coordinate fields.
		Only curves lying entirely within the block can be added.
		If any points have an invalid radius, whole curve is discarded.
		"""
		crvlist2 = []
		for crv in crvlist:
			t, r = crv.tr
			mask = np.logical_and( self.rj[0]<r , r<self.rj[1] )
			if not np.any(mask==False):
				crv.uv = self.uv_of_tr(crv.tr)
				crvlist2 += [crv]
		crvlist2 = self.update_curves_from_uv(crvlist2)
		return crvlist2

	def fill(self, sty={}):
		"""
		Fill in the block with a background patch.
		"""
		block_fill.fill_block(self, sty=sty)

	def fill_between_r(self, rvals=np.array([0.,0.]), sty={}, npoints=1000, inf=100):
		"""
		Fill block between r values if valid.
		"""
		block_fill.fill_between_r(self, rvals=rvals, sty=sty, npoints=npoints, inf=inf)


	def bplot(self):
		"""
		Plot all curves belonging to the block.
		"""
		self.refresh_curves_from_uv()
		for crv in self.curves:
			U, V = crv.UV[0], crv.UV[1]
			plt.plot(V-U, V+U, **crv.sty)

	"""
	Methods called by other methods.
	"""

	def update_derived_attributes(self):
		"""
		Use the current block primary attributes to set the derived attributes.
		"""
		## from metfunc
		self.rj = self.master.metfunc.rj[self.j:self.j+2]
		self.kj = self.master.metfunc.kj[self.j:self.j+2]
		self.kpm = np.sort( self.kj )
		self.sgnf = self.master.metfunc.sgnf(self.j)
		self.F    = lambda r: self.master.metfunc.F(r)
		self.Finv = lambda rstar: self.master.metfunc.Finv(self.j, rstar)
		## from region
		self.s0 = float(self.master.rparams['s0'])
		self.c  = float(self.master.rparams['c'])

	def update_coord_transf(self):
		"""
		Use the current block attributes to set the coordinate transformations.
		"""
		## extract block parameters from dict
		cdlu, cdlv = float(self.bparams['cdlu']), float(self.bparams['cdlv'])
		epsu, epsv = float(self.bparams['epsu']), float(self.bparams['epsv'])
		## define coord transformations with current parameters
		self.uv_of_tr   = lambda tr: coord.uv_of_tr(tr, self.F, self.c)
		self.tr_of_uv   = lambda uv: coord.tr_of_uv(uv, self.Finv, self.c)
		self.r_of_uv    = lambda uv: coord.r_of_uv(uv, self.Finv, self.c)
		self.uvdl_of_uv = lambda uv:   coord.uvdl_of_uv(uv  , cdlu, cdlv, epsu, epsv, self.kpm, self.s0)
		self.uv_of_uvdl = lambda uvdl: coord.uv_of_uvdl(uvdl, cdlu, cdlv, epsu, epsv, self.kpm, self.s0)
		## define supplementary functions for reference
		self.h    = lambda s: coord.hks(s, self.kpm, self.s0)
		self.hinv = lambda s: coord.hksinv(s, self.kpm, self.s0)

	def consistency_check(self):
		"""
		Check if current block attributes are self-consistent.
		True if consistent. False if not.
		"""
		## initialize
		consistent = True
		messages = [
			"WARNING: Block attributes are not self-consistent.",
			"Block Info: %s"%(self.master.metfunc.info),
			"Block Info: j = %r"%(self.j),
			"Issues:",
			]
		## # valid interval label?
		if not self.j in range(len(self.master.metfunc.rj)-1):
			consistent = False
			messages += ["Check block index.", "j=%r, jmax=%r"%( self.j, (len(self.master.metfunc.rj)-2) )]
		## # valid block orientation?
		epsu, epsv = float(self.bparams['epsu']), float(self.bparams['epsv'])
		if not epsu*epsv == self.sgnf:
			consistent = False
			messages += ["Check block orientation.", "epsu*epsv.sgnf = %r"%( epsu*epsv/self.sgnf )]
		## # warn if not consistent
		if consistent == False:
			messages += ["END WARNING"]
			print('\n' + '\n'.join(messages) + '\n')
		## return
		return consistent

	def update_curves_from_uv(self, crvlist):
		"""
		Update curves in the block, using uv coords to fill all other coordinate fields.
		"""
		for crv in crvlist:
			crv.r    = self.r_of_uv(crv.uv)
			crv.tr   = self.tr_of_uv(crv.uv)
			crv.uvdl = self.uvdl_of_uv(crv.uv)
			crv.UV   = self.master.UV_of_uvdl(crv.uvdl)
		return crvlist

	def apply_masks(self, crvlist):
		"""
		Apply all masks to self.curves.
		Masks are functions of the form 
			mask(block, crvlist): return new_crvlist
		"""
		for mask in self.masks:
			crvlist = mask(self, crvlist)
		return crvlist














"""
Tests for evap module.
"""

import numpy as np
import matplotlib.pyplot as plt
import xhorizon as xh
import pprint
import datetime

from xhorizon.evap.evap import *
from xhorizon.evap.helpers import *


def main():
	#test1()
	#test2()
	#test3()
	#test4()
	#test5()
	#test6()
	#test7()
	#test8()
	#test9()
	#test10()
	#test11()
	#test12()
	#test13()
	#test14()
	#test15()
	#test16()
	pass



def test1():
	"""
	Test functionality of get_rinf_uv0.
	"""
	##
	print("\nTEST 1\n")
	## regions
	reg1 = xh.reg.EFreg(xh.mf.schwarzschild())
	reg2 = xh.reg.EFreg(xh.mf.hayward())
	reg3 = xh.reg.EFreg(xh.mf.minkowski())
	reglist = [reg1,reg2,reg3]
	## v values
	v0 = [-1.,1.,0.]
	## get
	rinf = get_rinf_uv0(reglist,v0=v0)
	## print
	print(rinf)
	##
	print("\nEND TEST 1\n")



def test2():
	"""
	Test functionality of accrete.
	"""
	##
	print("\nTEST 2\n")
	## create initial region
	reglist = [xh.reg.EFreg(xh.mf.minkowski(),rlines=False,boundary=False)]
	## create accreted regions
	reglist += xh.evap.accrete(reglist.pop(), v1=0., v2=0., R2=0.8)
	reglist += xh.evap.accrete(reglist.pop(), v1=.5, v2=.5, R2=.9)
	reglist += xh.evap.accrete(reglist.pop(), v1=1., v2=1., R2=1.)
	## add lines
	xh.evap.colorlines(reglist)
	xh.evap.boundarylines(reglist)
	## draw
	plt.figure()
	plt.gca().set_aspect('equal')
	for reg in reglist:
		reg.rplot()
	plt.show()
	##
	print("\nEND TEST 2\n")



def test3():
	"""
	Test functionality of accretion.
	"""
	##
	print("\nTEST 3\n")
	## params
	v = np.linspace(0,1,5)
	m = .2 + .8*v
	## create initial region
	reglist = accretion(m,v)
	## add lines
	xh.evap.colorlines(reglist)
	xh.evap.boundarylines(reglist)
	## draw
	plt.figure()
	plt.gca().set_aspect('equal')
	for reg in reglist:
		reg.rplot()
	plt.show()
	##
	print("\nEND TEST 3\n")





def test4():
	"""
	Test functionality of evap.
	"""
	##
	print("\nTEST 4\n")
	## create initial region
	reg0 = xh.reg.EFreg(xh.mf.hayward(R=1.,l=0.1),rlines=False,boundary=False)
	reglist = [reg0]
	## create evaporated regions
	reglist += xh.evap.evap(reglist.pop(), u1=10., v1=-10., u2=10., R2=0.9)
	#reglist += xh.evap.evap(reglist.pop(), u1=2., v1=2., u2=0., R2=0.98)
	## check
	check_uvr(reglist)
	## draw diagram?
	if True:
		## add lines
		xh.evap.colorlines(reglist, sty=dict(marker='.', markersize=.3, ls='none'), rmin=0.8, rmax=1.1, dr=.02, npoints=2001, inf=25.)
		xh.evap.boundarylines(reglist)
		xh.evap.s0_lines(reglist)
		## draw
		xh.newfig(tex=False,sqaxis=3)
		plt.title("Test 4")
		for reg in reglist:
			reg.rplot()
		## fill
		fill_by_R(reglist)
		## show plot
		plt.savefig("temp-figs/test4.png", dpi=400)
		##
	print("\nEND TEST 4\n")


def test5():
	"""
	Test functionality of evaporation().
	"""
	##
	print("\nTEST 5\n")
	## create initial region
	reg0 = xh.reg.EFreg(xh.mf.hayward(R=1.,l=0.1),rlines=False,boundary=False)
	reglist = [reg0]
	## init params
	u0 = 10.
	v0 = 0.
	## subsequent params
	Nreg = 1
	tt = np.linspace(0.,1.,Nreg+2)[1:-1]
	R = 1. * (1.-tt)**(1./3.)
	du = 0.4 + 0.*tt
	dv = 0.2 + 0.*tt
	## create evaporated regions
	reglist += xh.evap.evaporation(reglist.pop(), R=R, du=du, dv=dv, u0=u0, v0=v0, l=0.1, rparams={})
	## check uvr
	check_uvr(reglist)
	## draw regions?
	if True:
		## add lines
		xh.evap.colorlines(reglist)
		xh.evap.boundarylines(reglist)
		## draw
		xh.newfig(tex=False,sqaxis=3)
		plt.title('Test 5')
		for reg in reglist:
			reg.rplot()
		## fill
		fill_by_R(reglist, cm=plt.cm.prism)
		## show plot
		plt.savefig("temp-figs/test5.png", dpi=200)
		##
	print("\nEND TEST 5\n")
	## return
	return reglist


def test6():
	"""
	Test functionality of formevap().
	"""
	##
	print("\nTEST 6\n")
	## accrete params
	N = 4.
	R0, R = .8, 1.
	accrete_R = R0 + (R-R0)*np.linspace(0,1,N)
	v0, v = -1., -.5
	accrete_v = np.linspace(v0,v,N)
	## evap params
	evap_R = [.9 ,.0]
	evap_u = [6.,9.]
	## create evaporated regions
	reglist = xh.evap.formevap(accrete_R=accrete_R, accrete_v=accrete_v, evap_R=evap_R, evap_u=evap_u, l=0.01, rparams=dict(c=0., s0=10.))
	## resquish
	fU = None
	fV = None
	reglist = UVcompose(reglist, fU=fU, fV=fV)
	## add lines
	xh.evap.colorlines(reglist)
	xh.evap.boundarylines(reglist)
	## draw
	xh.newfig(tex=False,sqaxis=3)
	#plt.figure()
	plt.title('Test 6')
	plt.xlim(0.5,2.5)
	plt.ylim(-1,1)
	for reg in reglist:
		reg.rplot()
	## fill
	fill_by_R(reglist)
	## mink compare
	#xh.reg.EFreg(xh.mf.minkowski()).rplot()
	## show plot
	plt.savefig("temp-figs/test6.png", dpi=600)
	##
	print("\nEND TEST 6\n")


def test7():
	"""
	Test functionality of get_uvdl_of_UV().
	"""
	##
	print("\nTEST 7\n")
	## create evaporated regions
	reglist = xh.evap.formevap(rparams=dict(s0=10.))
	## resquish
	fU, fV = get_uvdl_of_UV(reglist[-1])
	## plot
	xx = np.linspace(-2,2,50001)
	plt.figure()
	plt.plot(xx, fU(xx), 'rx-')
	plt.plot(xx, fV(xx), 'bx-')
	plt.grid()
	plt.show()
	## edit regions
	UVcompose(reglist, fU=fU, fV=fV)
	## add lines
	xh.evap.colorlines(reglist)
	xh.evap.boundarylines(reglist)
	## draw
	xh.newfig(tex=False,sqaxis=3)
	plt.title('Test 7')
	for reg in reglist:
		reg.rplot()
	fill_by_R(reglist)
	## show plot
	plt.show()
	plt.savefig("temp-figs/test7.png", dpi=400)
	##
	print("\nEND TEST 7\n")



def test8():
	"""
	Test functionality of EF cornermask.
	"""
	##
	print("\nTEST 8\n")
	## make region
	func = xh.mf.schwarzschild()
	reg = xh.reg.EFreg(func)
	##
	if True:
		abcd = 'a'
		reg = xh.cornermask.EFreg(reg, abcd=abcd, u0=0., v0=0.)
	##
	plt.figure()
	plt.xlim(-3,3)
	plt.gca().set_aspect('equal')
	reg.rplot()
	plt.show()
	##
	print("\nEND TEST 8\n")




def test9():
	"""
	Test functionality of split abcd helper.
	"""
	##
	print("\nTEST 9\n")
	## make region
	func = xh.mf.schwarzschild()
	reg = xh.reg.EFreg(func)
	##
	if True:
		abcd = 'bcd'
		reglist = xh.evap.split_reg_abcd(reg, abcd=abcd, u0=0., v0=0.)
	##
	print(len(reglist))
	print(reglist)
	##
	plt.figure()
	plt.xlim(-3,3)
	plt.gca().set_aspect('equal')
	for reg in reglist:
		reg.rplot()
	plt.show()
	##
	print("\nEND TEST 9\n")



def test10():
	"""
	Test direct formation of joint. Formerly was evap.debug.go().
	"""
	##
	print("\nTEST 10\n")
	for ux in [15.]:
		## params
		R1 = 1.
		R2 = .5
		l = .1
		rparams = dict(s0=10.)
		u1 = ux
		v1 = 0.
		u2 = 1.*u1
		## funcs
		if False:
			func1 = xh.mf.hayward(R=R1,l=l)
			func2 = xh.mf.hayward(R=R2,l=l)
		else:
			func1 = xh.mf.schwarzschild(R=R1)
			func2 = xh.mf.schwarzschild(R=R2)
		## regs
		reg1 = xh.reg.EFreg(func1,rparams=rparams,rlines=False,boundary=False)
		reg2 = xh.reg.EFreg(func2,rparams=rparams,rlines=False,boundary=False)
		## horizon radii
		Rh1, Rh2 = reg1.blocks[-1].rj[0], reg2.blocks[-1].rj[0]
		## passive slice of reg1
		r1 = reg1.blocks[-1].r_of_uv(np.array([[u1],[v1]]))[0]
		pslice = xh.junc.pslice(reg1, ublocks=[-1], vblocks=range(len(reg1.blocks)), r0=r1, u0=u1)
		## active slice of reg2
		aslice = xh.junc.aslice(reg2, ublocks=[-1], vblocks=range(len(reg2.blocks)), r0=pslice.r0, u0=u2, U0=pslice.U_of_r_at_v0, V0=pslice.V_of_r_at_u0, r_refs=[pslice.reg.metfunc.r_ref])
		## plot slice
		if False:
			print('REG1 PSLICE U(r,v0)=red, V(r,u0)=blue')
			print('REG2 ASLICE U(r,v0)=red, V(r,u0)=blue')
			plt.figure()
			plt.title('REG1 PSLICE U(r,v0)=red, V(r,u0)=blue\nREG2 ASLICE U(r,v0)=mag, V(r,u0)=cyan')
			plt.xlabel('r')
			plt.ylabel('U(r),V(r) on slice')
			plt.grid()
			plt.xlim(.5,1.5)
			## passive slice
			style1 = dict(marker='x', markersize=8, lw=1, ls='none')
			plt.plot(pslice.r, pslice.UV_v0[0], c='r', **style1)
			plt.plot(pslice.r, pslice.UV_u0[1], c='b', **style1)
			## active slice
			style2 = dict(marker='o', markersize=6, lw=1, ls='none')
			plt.plot(aslice.r, aslice.UV_v0[0], c='m', **style2)
			plt.plot(aslice.r, aslice.UV_u0[1], c='c', **style2)
			## junction radius
			plt.plot([pslice.r0, pslice.r0], [-1.,1.], 'k-' )
		## print slice params
		print("\n")
		print("                          %22r, %22r, %22r, %22r, %22r"%('Rh', 'r', 'r/Rh', 'u', 'v'))
		print("Pslice Inputs:")
		print("Rh1, r1, r1/Rh1, u1, v1 = %22r, %22r, %22r, %22r, %22r"%(Rh1, r1, r1/Rh1, u1, v1))
		print("From PSlice:")
		print("Rh1, r1, r1/Rh1, u1, v1 = %22r, %22r, %22r, %22r, %22r"%(Rh1, pslice.r0, pslice.r0/Rh1, pslice.u0, pslice.v0))
		print("Aslice Inputs:")
		print("Rh2, r2, r2/Rh2, u2, v2 = %22r, %22r, %22r, %22r, %22r"%(Rh2, pslice.r0, pslice.r0/Rh2, u2, '?'))
		print("From ASlice:")
		print("Rh2, r2, r2/Rh2, u2, v2 = %22r, %22r, %22r, %22r, %22r"%(Rh2, aslice.r0, aslice.r0/Rh2, aslice.u0, aslice.v0))
		print("\n")
		## reglist
		reglist = []
		## split regions
		reglist1 = xh.evap.split_reg_abcd(reg1,  abcd='bcd', u0=pslice.u0, v0=pslice.v0)
		reglist2 = xh.evap.split_reg_abcd(reg2,  abcd='a',   u0=aslice.u0, v0=aslice.v0)
		## update transformations for reg2
		for i in range(len(reglist2)):
			reglist2[i].U_of_udl = aslice.U_of_udl_at_v0
			reglist2[i].V_of_vdl = aslice.V_of_vdl_at_u0
		## add to reglist
		reglist += reglist1
		reglist += reglist2
		## print block uvbounds
		print("\n")
		print('%22r, %22r, %22r, %22r, %22r, %22r'%('R', 'b.rj', 'vmin', 'vmax', 'umin', 'umax'))
		for i in range(len(reglist)):
			reg = reglist[i]
			for b in reg.blocks:
					print('%22s, %22s, %22r, %22r, %22r, %22r'%(reg.metfunc.fparams['R'], b.rj, b.uvbounds['vmin'], b.uvbounds['vmax'], b.uvbounds['umin'], b.uvbounds['umax']))
		print("\n")
		## draw diagram
		if True:
			## add lines
			xh.evap.colorlines(reglist, sty=dict())
			xh.evap.boundarylines(reglist, sty=dict(), npoints=5001)
			xh.evap.s0_lines(reglist, sty=dict(lw=3))
			## draw
			plt.figure()
			plt.xlim(-3,3)
			plt.gca().set_aspect('equal')
			for reg in reglist:
				reg.rplot()
			## fill
			xh.evap.fill_by_R(reglist, cm=plt.cm.prism)
	## show plots
	plt.show()
	##
	print("\nEND TEST 10\n")






def test11(u0=-12., v0=0., du=.5, dv=.3, R=np.array([1.,.98,.96])):
	"""
	One intermediate region.
	Build from bottom.
	"""
	##
	print("\nTEST 11\n")
	## reglist
	reglist = []
	## funcs
	funclist = [xh.mf.schwarzschild(R=Rx) for Rx in R]
	## first region
	reglist += [xh.reg.EFreg(funclist.pop(0), boundary=False, rlines=False)]
	## second region
	reg1 = reglist.pop()
	reg2 = xh.reg.EFreg(funclist.pop(0), boundary=False, rlines=False)
	u0 = 1.*u0
	v0 = 1.*v0
	r0 = reg1.blocks[-1].r_of_uv(np.array([[u0],[v0]]))[0]
	pslice = xh.junc.pslice(reg1, ublocks=[-1], vblocks=range(len(reg1.blocks)), r0=1.*r0, u0=1.*u0)
	aslice = xh.junc.aslice(reg2, ublocks=[-1], vblocks=range(len(reg2.blocks)), r0=1.*pslice.r0, v0=1.*pslice.v0, U0=pslice.U_of_r_at_v0, V0=pslice.V_of_r_at_u0, r_refs=[pslice.reg.metfunc.r_ref])
	reg2.U_of_udl = aslice.U_of_udl_at_v0
	reg2.V_of_vdl = aslice.V_of_vdl_at_u0
	reglist += xh.evap.split_reg_abcd(reg1,  abcd='dbc', u0=1.*pslice.u0, v0=1.*pslice.v0)
	reglist += xh.evap.split_reg_abcd(reg2,  abcd='a'  , u0=1.*aslice.u0, v0=1.*aslice.v0)
	sliceprint(pslice, aslice, reg1.blocks[-1].rj[0], reg2.blocks[-1].rj[0])
	## third region
	reg1 = reglist.pop()
	reg2 = xh.reg.EFreg(funclist.pop(0), boundary=False, rlines=False)
	u0 = 1.*aslice.u0 + du
	v0 = 1.*aslice.v0 + dv
	r0 = reg1.blocks[-1].r_of_uv(np.array([[u0],[v0]]))[0]
	pslice = xh.junc.pslice(reg1, ublocks=[-1], vblocks=range(len(reg1.blocks)), r0=1.*r0, u0=1.*u0)
	aslice = xh.junc.aslice(reg2, ublocks=[-1], vblocks=range(len(reg2.blocks)), r0=1.*pslice.r0, v0=1.*pslice.v0, U0=pslice.U_of_r_at_v0, V0=pslice.V_of_r_at_u0, r_refs=[pslice.reg.metfunc.r_ref])
	reg2.U_of_udl = aslice.U_of_udl_at_v0
	reg2.V_of_vdl = aslice.V_of_vdl_at_u0
	reglist += xh.evap.split_reg_abcd(reg1,  abcd='dbc', u0=1.*pslice.u0, v0=1.*pslice.v0)
	reglist += xh.evap.split_reg_abcd(reg2,  abcd='a'  , u0=1.*aslice.u0, v0=1.*aslice.v0)
	sliceprint(pslice, aslice, reg1.blocks[-1].rj[0], reg2.blocks[-1].rj[0])
	## plot
	rgp(reglist)
	## show
	plt.show()
	##
	print("\nEND TEST 11\n")




def test12(u0=12., dr0=.1, du=.5, R=np.array([1.,.98,.96])):
	"""
	One intermediate region.
	Build from top.
	"""
	##
	print("\nTEST 12\n")
	## reglist
	reglist = []
	## funcs
	funclist = [xh.mf.schwarzschild(R=Rx) for Rx in R]
	## first region
	reglist += [xh.reg.EFreg(funclist.pop(-1), boundary=False, rlines=False)]
	## second region
	reg1 = reglist.pop()
	reg2 = xh.reg.EFreg(funclist.pop(-1), boundary=False, rlines=False)
	u0 = 1.*u0
	r0 = 1.*reg2.blocks[-1].rj[0] + 1.*dr0
	pslice = xh.junc.pslice(reg1, ublocks=[-1], vblocks=range(len(reg1.blocks)), r0=1.*r0, u0=1.*u0)
	aslice = xh.junc.aslice(reg2, ublocks=[-1], vblocks=range(len(reg2.blocks)), r0=1.*pslice.r0, u0=1.*pslice.u0, U0=pslice.U_of_r_at_v0, V0=pslice.V_of_r_at_u0, r_refs=[pslice.reg.metfunc.r_ref])
	reg2.U_of_udl = aslice.U_of_udl_at_v0
	reg2.V_of_vdl = aslice.V_of_vdl_at_u0
	reglist += xh.evap.split_reg_abcd(reg1,  abcd='a', u0=1.*pslice.u0, v0=1.*pslice.v0)
	reglist += xh.evap.split_reg_abcd(reg2,  abcd='dbc'  , u0=1.*aslice.u0, v0=1.*aslice.v0)
	sliceprint(pslice, aslice, reg1.blocks[-1].rj[0], reg2.blocks[-1].rj[0])
	## third region
	reg3 = xh.reg.EFreg(funclist.pop(-1), boundary=False, rlines=False)
	u0 = 1.*aslice.u0 - 1.*du
	r0 = 1.*reg3.blocks[-1].rj[0] + 1.*dr0
	pslice = xh.junc.pslice(reg2, ublocks=[-1], vblocks=range(len(reg1.blocks)), r0=1.*r0, u0=1.*u0)
	aslice = xh.junc.aslice(reg3, ublocks=[-1], vblocks=range(len(reg2.blocks)), r0=1.*pslice.r0, u0=1.*pslice.u0, U0=pslice.U_of_r_at_v0, V0=pslice.V_of_r_at_u0, r_refs=[pslice.reg.metfunc.r_ref])
	reg3.U_of_udl = aslice.U_of_udl_at_v0
	reg3.V_of_vdl = aslice.V_of_vdl_at_u0
	for reg in reglist[-3:-1]:
		for b in reg.blocks:
			b.uvbounds.update(dict(vmin=1.*pslice.v0))
	for reg in reglist[-2:]:
		for b in reg.blocks:
			b.uvbounds.update(dict(umin=1.*pslice.u0))
	reglist += xh.evap.split_reg_abcd(reg3,  abcd='dbc' , u0=1.*aslice.u0, v0=1.*aslice.v0)
	sliceprint(pslice, aslice, reg2.blocks[-1].rj[0], reg3.blocks[-1].rj[0])
	## plot
	rgp(reglist)
	## show
	plt.show()
	##
	print("\nEND TEST 12\n")





def test13():
	"""
	Test evap.evap.funclist_chain().
	"""
	##
	print("\nTEST 13\n")
	## make funclist
	R=np.array([0.,.5,1.,.7,0.])
	funclist  = xh.evap.formevap_funclist(R=R)
	## init
	t0 = np.zeros(len(funclist))
	x0 = np.zeros(len(funclist))
	## params
	t0 = -2. + t0
	x0 = 2. + x0
	## params
	## params
	ss = np.ones(len(funclist))
	du  = 1.*ss
	dv  = .9*ss
	##
	reglist, chainparams = xh.evap.funclist_chain(funclist, seed=-2, u0=t0-x0, v0=t0+x0, du=1.*du, dv=1.*dv, eta=1., matchmode='rv')
	pprint.pprint(chainparams)
	##
	reglist, chainparams = xh.evap.chain_masker(reglist, chainparams)
	pprint.pprint(chainparams)
	##
	rgp(reglist)
	plt.show()
	##
	print("\nEND TEST 13\n")




def test14():
	"""
	Test evap.evap.shellparams_list().
	"""
	##
	print("\nTEST 14\n")
	##
	## shellparams
	sp = xh.evap.shellparams_list(Rmin=.2, Rmax=1., dv=1., l=.1, A=10., functype=xh.mf.schwarzschild, fparams=dict())
	spt = xh.evap.sp_transpose(sp)
	## outs
	RR = spt['R']
	du = spt['du']
	dv = spt['dv']
	uu = spt['uu']
	vv = spt['vv']
	A = spt['A'][0]
	## smooth uu,vv
	ss = np.linspace(0,1.1*np.max(uu),5001)
	## plot
	if True:
		plt.figure()
		plt.title('R vs u')
		plt.xlabel('u')
		plt.ylabel('A*R^3')
		plt.xlim(0,10)
		plt.ylim(0,10)
		plt.plot(uu,A*RR**3,'k.')
		plt.plot(ss,ss,'b-')
		plt.grid(1)
	## plot
	if False:
		plt.figure()
		plt.title('R vs v')
		plt.xlabel('v')
		plt.ylabel('A*R^3')
		plt.ylim(0,A)
		plt.plot(vv,A*RR**3,'k.')
		plt.grid(1)
	## plot
	if False:
		plt.figure()
		plt.title('du/dv vs v')
		plt.xlabel('v')
		plt.ylabel('du/dv')
		plt.plot(vv,du/dv,'k.')
		plt.grid(1)
	## show
	plt.show()
	##
	##
	print("\nEND TEST 14\n")




def test15():
	"""
	Test evap.evap.shellparams_list().
	"""
	##
	print("\nTEST 15\n")
	##
	## shellparams
	sp = xh.evap.shellparams_list(Rmin=.2, Rmax=1., dv=1., l=.1, A=1., functype=xh.mf.schwarzschild, fparams=dict())
	spt = xh.evap.sp_transpose(sp)
	## outs
	funclist = spt['funclist']
	du = spt['du']
	dv = spt['dv']
	r0f = spt['R'] + spt['l']
	r0p = np.roll(r0f,1)
	v0 = 0.*np.linspace(-4, 4, len(funclist))
	##
	print("chain")
	reglist, chainparams = xh.evap.funclist_chain(funclist, seed=0, du=1.*du, dv=1.*dv, r0p=1.*r0p, r0f=1.*r0f, v0=1.*v0)
	pprint.pprint(chainparams)
	##
	print("mask")
	reglist, chainparams = xh.evap.chain_masker(reglist, chainparams)
	pprint.pprint(chainparams)
	##
	print("plot")
	rgp(reglist)
	plt.savefig('temp-figs/test15.png', dpi=200)
	##
	print("\nEND TEST 15\n")



def test16():
	"""
	Test evap.evap.formevap_input() with funclist_chain().
	"""
	##
	print("\nTEST 16\n")
	## params
	params = dict()
	## funcs
	params.update(dict(functype0=xh.mf.minkowski, fparams0=dict(), functype1=xh.mf.hayward, fparams1=dict(l=.01)))
	## evap
	params.update(dict(Rmin=1., Rmax=1., dv_evap=.5, l=.1, A=.2))
	## accrete
	params.update(dict(B=.5, Naccrete=1))
	## offset
	params.update(dict(voff=0., veta=1., uoff=0., ueta=0.))
	## seed
	seed = 0
	##
	print("inputs")
	funclist, cp = formevap_input(**params)
	##
	print("chain")
	reglist, chainparams = funclist_chain(funclist, seed=seed, **cp)
	##
	print("mask")
	reglist, chainparams = xh.evap.chain_masker(reglist, chainparams)
	pprint.pprint(chainparams)
	##
	print("plot")
	rgp(reglist[:])
	plt.xlim(0,3)
	plt.ylim(-1.5,1.5)
	##
	print("save fig")
	fname =  datetime.datetime.now().strftime("Test16_%Y-%m-%d_%H-%M-%S-%f")
	plt.title(fname)
	plt.savefig("temp-figs/%s.png"%(fname), dpi=400)
	##
	print("save txt")
	ff = open("temp-figs/%s.txt"%(fname), 'w')
	ff.write("%s\n"%(fname))
	ff.write('\n')
	ff.write('Input:\nparams=\n%s\nseed=\n%s\n'%(pprint.pformat(params),seed))
	ff.write('\n')
	ff.write('Output:\nchainparams=\n%s\n'%(pprint.pformat(chainparams)))
	##
	print("\nEND TEST 16\n")


main()





from xhorizon.evap import schwarzparams
from xhorizon.evap import helpers
from xhorizon.evap import evap

from xhorizon.evap import draw
from xhorizon.evap import draw_dS

from xhorizon.evap import demo
from xhorizon.evap import demo_AdS
from xhorizon.evap import demo_dS
from xhorizon.evap import demo_zoom
from xhorizon.evap import massplot
from xhorizon.evap import legend
from xhorizon.evap import tests

from xhorizon.evap.evap import *
from xhorizon.evap.draw import drawreg
from xhorizon.evap.demo import demo
from xhorizon.evap import schwarzparams as SSp





"""
This module provides functions for applying uvbounds masks to correct blocks in a given region.
These functions should directly edit and return the input region object.
"""

import numpy as np


def EFreg(reg, abcd=None, u0=None, v0=None):
	"""
	"""
	## sort
	if (len(reg.blocks)-1)==1:
		if reg.metfunc.sgnf(0)==-1.0:
			out = EFreg1a(reg, abcd=abcd, u0=u0, v0=v0)
	if (len(reg.blocks)-1)==2:
		if reg.metfunc.sgnf(0)==1.0:
			out = EFreg2a(reg, abcd=abcd, u0=u0, v0=v0)
	## return
	return out




def EFreg2a(reg, abcd=None, u0=None, v0=None):
	"""
	For Hayward-like regions (aka N=2 and f(0)>0).
	Assumes slice in outermost block.
		reg = The region to mask.
		abcd = Which corner junction label is this piece, with standard label scheme. 
			Should be a single letter string 'a' 'b' 'c' or 'd'.
		u0, v0 = Location to slice.
	Returns the same region that was input, after editing.

	Region is masked by removing unwanted blocks and setting uvbounds on remaining blocks.

	Block order is [inner, middle, outer] (this is arbitrary, just happens to be how EFreg2a is generated).
	Setting any of the uvbounds to np.nan causes all inequalities to fail, which causes an error unless block is removed.
	"""
	## initialize
	uvb = [dict(), dict(), dict()]
	## set proper uvbounds for each block based on abcd type
	## case a
	if abcd=='a':
		uvb = [dict(vmin=v0), dict(vmin=v0), dict(vmin=v0, umin=u0)]
	## case b
	if abcd=='b':
		uvb = [dict(vmin=np.nan), dict(vmin=np.nan), dict(vmax=v0,umax=u0)]
	## case c
	if abcd=='c':
		uvb = [dict(vmin=np.nan), dict(vmin=np.nan), dict(vmin=v0,umax=u0)]
	## case d
	if abcd=='d':
		uvb = [dict(vmax=v0), dict(vmax=v0), dict(vmax=v0, umin=u0)]
	## update blocks uvbounds
	for i in range(len(reg.blocks)):
		reg.blocks[i].uvbounds.update(uvb[i])
	## keep blocks only if no nan in uvbounds values
	keep = []
	for b in reg.blocks:
		if not np.nan in b.uvbounds.values():
			keep += [b]
	reg.blocks = keep
	## return
	return reg


def EFreg1a(reg, abcd=None, u0=None, v0=None):
	"""
	For Schwarzschild-like regions (aka N=1 and f(0)<0).
	Assumes slice in outermost block.
		reg = The region to mask.
		abcd = Which corner junction label is this piece, with standard label scheme. 
			Should be a single letter string 'a' 'b' 'c' or 'd'.
		u0, v0 = Location to slice.
	Returns the same region that was input, after editing.

	Region is masked by removing unwanted blocks and setting uvbounds on remaining blocks.

	Block order is [inner, outer] (this is arbitrary, just happens to be how EFreg1a is generated).
	Setting any of the uvbounds to np.nan causes all inequalities to fail, which causes an error unless block is removed.
	"""
	## initialize
	uvb = [dict(), dict()]
	## set proper uvbounds for each block based on abcd type
	## case a
	if abcd=='a':
		uvb = [dict(vmin=v0), dict(vmin=v0, umin=u0)]
	## case b
	if abcd=='b':
		uvb = [dict(vmin=np.nan), dict(vmax=v0,umax=u0)]
	## case c
	if abcd=='c':
		uvb = [dict(vmin=np.nan), dict(vmin=v0,umax=u0)]
	## case d
	if abcd=='d':
		uvb = [dict(vmax=v0), dict(vmax=v0, umin=u0)]
	## update blocks uvbounds
	for i in range(len(reg.blocks)):
		reg.blocks[i].uvbounds.update(uvb[i])
	## keep blocks only if no nan in uvbounds values
	keep = []
	for b in reg.blocks:
		if not np.nan in b.uvbounds.values():
			keep += [b]
	reg.blocks = keep
	## return
	return reg



def MAXreg2a(reg, abcd=None, u0=None, v0=None):
	"""
	For Schwarzschild-like regions (aka N=2 and f(0)>0).
		reg = The region to mask.
		abcd = Which corner junction label is this piece, with standard label scheme. 
			Should be a single letter string 'a' 'b' 'c' or 'd'.
		u0, v0 = Location to slice.
	Returns the same region that was input, after editing.

	Region is masked by removing unwanted blocks and setting uvbounds on remaining blocks.

	Block order is [top, right, bottom, left] (this is arbitrary, just happens to be how MAXreg2a is generated).
	Setting any of the uvbounds to np.nan causes all inequalities to fail, which causes an error unless block is removed.
	"""
	## initialize
	uvb = [dict(), dict(), dict(), dict()]
	## set proper uvbounds for each block based on abcd type
	## case a
	if abcd=='a':
		uvb = [dict(vmin=v0), dict(vmin=v0,umin=u0), dict(vmin=np.nan), dict(vmin=np.nan)]
	## case b
	if abcd=='b':
		uvb = [dict(vmin=np.nan), dict(vmax=v0,umax=u0), dict(umax=u0), dict(vmin=np.nan)]
	## case c
	if abcd=='c':
		uvb = [dict(vmin=np.nan), dict(vmin=v0,umax=u0), dict(vmin=np.nan), dict(vmin=np.nan)]
	## case d
	if abcd=='d':
		uvb = [dict(vmax=v0), dict(vmax=v0,umin=u0), dict(umin=u0), dict()]
	## update blocks uvbounds
	for i in range(len(reg.blocks)):
		reg.blocks[i].uvbounds.update(uvb[i])
	## keep blocks only if no nan in uvbounds values
	keep = []
	for b in reg.blocks:
		if not np.nan in b.uvbounds.values():
			keep += [b]
	reg.blocks = keep
	## return
	return reg












"""
This module defines the classes
	curve
which is used to construct and plot diagrams.
"""

import numpy as np

class curve:

	"""
	A diagram is fundamentally a collection of curves.
	A curve is sequence of points in spacetime, which can be defined in many coordinate systems.
	Every curve must at least have diagram coordinates UV for plotting, but not all coordinate systems must be filled.
	Each point has a well-defined radius independent of coordinate system.
	Curve attributes:
		r = radius
		tr = t,r = schwarzschild block coordinates
		uv = u,v = double null block coordinates
		uvdl = \tilde{u},\tilde{v} = region coordinates (standard penrose coords on the sss region)
		UV = U,V = diagram coordinates
		sty = dictionary of style parameters
	"""

	def __init__(self):
		## radius
		self.r = np.array([])
		## block coordinates
		self.tr = np.array([[],[]])
		## block coordinates
		self.uv = np.array([[],[]])
		## region coordinates
		self.uvdl = np.array([[],[]])
		## diagram coordinates
		self.UV = np.array([[],[]])
		## style
		self.sty = dict(c='k', ls='-', lw=0.5)







import numpy as np
import matplotlib.pyplot as plt


"""
Consider an evaporating black hole with maximum mass M.
Suppose it evaporates by emitting Nevap shells of Hawking radiation.
This is described by a sequence of 
    Nevap+1 regions (first with m=M, last with m=0) with parameters m_j, du_j, dv_j, rp_j, rf_j
    Nevap shells (future boundary of each region) with parameters dmi, r_i, u_i, v_i
Suppose each is Schwarz so that
 	F(r) = 2m*( (r/2m-1) + log(r/2m-1) ).
Evap length is le.

j
M  ...      ... 0
0 | 1 | 2 | ...
  0   1   2  ...
  i

Constraints:
	R_j  = 2m_j
	dm_i = m_i+1 - m_i
	du_j = u_j - u_j-1
	dv_j = v_j - v_j-1

	dm_i < 0
	du_j = -3T/M m_j^2 dm_j-1 > 0
	dv_j - du_j = 2( F(rf_j) - F(rp_j) )

	rf_j = Rh(m_j)+le
	rp_j = Rh(m_j-1)+le

"""


def F(r,m):
	return r + 2.*m * np.log(np.abs(r/(2.*m) - 1))

def R(m):
	return 2.*m



def SSduvm(Nevap=1e4, Tevap=1e4, M=.5, le=1e-4):
	"""
	"""
	##
	eps = 1e-9
	s = np.linspace(eps,1.+eps,Nevap+1)[:-1]
	s = np.append(.5*eps,s)
	## u
	u  = Tevap * s
	## m(u)
	m = M * (1. - u/Tevap)**(1./3.)
	## r(u)
	r = 2.*m + le
	## v(u)
	v = u + 2.*F(r,m)
	## du dv dm
	du = u[1:]-u[:-1]
	dv = v[1:]-v[:-1]
	## mself mprev
	mself = m[1:]
	mprev = m[:-1]
	## drstar
	drstar = F(R(mself)+le,mself) - F(R(mprev)+le,mself)
	## zero
	zero = dv-du-2.*drstar
	## print 
	print("|dv-du-2.*drstar| (mean) = %s"%(np.mean(np.abs(zero))))
	print("|dv-du-2.*drstar| (total) = %s"%(Nevap*np.mean(np.abs(zero))))
	## return
	return mself, du, dv


#SSduvm()






"""
This module provides method for making forming and evaporation BH diagrams.
This module imports the entire xhorizon package. 
It is meant for a higher level usage than the other subpackages, none of the
guts of xhorizon rely on this.
"""

import numpy as np
import matplotlib.pyplot as plt
import copy, pprint
import scipy.optimize as opt

import xhorizon as xh
from xhorizon.evap.helpers import *



###############################################################################################################3



def funclist_chain(funclist, seed=0, du=None, dv=None, r0p=None, r0f=None, u0=None, v0=None, ps_matchmode=None, fs_matchmode=None):
	"""
	Create a chain of matched regions, starting at seed region which is unmodified.
	Each region except ends has two slices through it, a future slice fslice and past slice pslice.
	Each fslice and pslice can be either active or passive, but there can only be one active slice per region.
	The index i refers to each region in the sequence for all variables.

	Inputs:
		funclist = list of func objects, in order, to chain together
		seed = index value for seed region (seed region has trivial transforms to target coords)
		du = list of du values so that du[i] will always be size of region[i]
		dv = list of du values so that du[i] will always be size of region[i]
		r0p = list of r0 values for past slice so that r0p will always be ps_r0 when pslice is active
		r0f = list of r0 values for future slice so that r0f will always be fs_r0 when fslice is active
		u0 = list of offset values for range of u values in slice, defaults to zero
		v0 = list of offset values for range of v values in slice, defaults to zero
		ps_matchmode = list of strings, each either 'ru' or 'rv', to determine how past slice is sliced when pslice is active
		ps_matchmode = list of strings, each either 'ru' or 'rv', to determine how future slice is sliced when fslice is active
	"""
	print("du funclist_chain")
	print(repr(du))
	print("dv funclist_chain")
	print(repr(dv))
	## init default values
	if u0==None:
		u0 = np.zeros(len(funclist))
	if v0==None:
		v0 = np.zeros(len(funclist))
	if ps_matchmode==None:
		ps_matchmode = ['rv' for func in funclist]
	if fs_matchmode==None:
		fs_matchmode = ['rv' for func in funclist]
	## set irrelevant first and last du and dv values to zero
	du[0], du[-1] = 0., 0.
	dv[0], dv[-1] = 0., 0.	
	## init internal variables
	reglist = [xh.reg.EFreg(funcx, boundary=False, rlines=False) for funcx in funclist]
	pslice  = [None for funcx in funclist]
	fslice  = [None for funcx in funclist]
	Rh      = [funcx.rj[-2] for funcx in funclist]
	ps_r0   = [np.nan for funcx in funclist]
	ps_u0   = [np.nan for funcx in funclist]
	ps_v0   = [np.nan for funcx in funclist]
	fs_r0   = [np.nan for funcx in funclist]
	fs_u0   = [np.nan for funcx in funclist]
	fs_v0   = [np.nan for funcx in funclist]
	i0 = range(len(funclist))[1*seed]
	ps_matchpop = [mp(mmm) for mmm in ps_matchmode]
	fs_matchpop = [mp(mmm) for mmm in fs_matchmode]
	## seed region
	i = 1*i0
	for i in [1*i0]:
		###### past passive slice
		## past passive slice input params (not mutually consistent)
		ps_u0[i]  = u0[i] - 0.5*du[i]
		ps_v0[i]  = v0[i] - 0.5*dv[i]
		ps_r0[i]  = 1.*r0p[i]
		## get past passive slice location from inputs and matchpop
		sliceloc = dict(u0=ps_u0[i], v0=ps_v0[i], r0=ps_r0[i])
		sliceloc.pop(ps_matchpop[i])
		print("i=%s pslice loc: %s"%(i,sliceloc))
		## execute past passive slice at sliceloc
		pslice[i] = xh.junc.pslice(reglist[i], ublocks=[-1], vblocks=range(len(reglist[i].blocks)), **sliceloc)
		## update past passive slice location to true values
		ps_u0[i], ps_v0[i], ps_r0[i]  = 1.*pslice[i].u0, 1.*pslice[i].v0, 1.*pslice[i].r0
		#### future passive slice
		## future passive slice input params (not mutually consistent)
		fs_u0[i]  = 1.*ps_u0[i] + 1.*du[i]
		fs_v0[i]  = 1.*ps_v0[i] + 1.*dv[i]
		fs_r0[i]  = 1.*r0f[i]
		## get future passive slice location from inputs and matchpop
		sliceloc = dict(u0=fs_u0[i], v0=fs_v0[i], r0=fs_r0[i])
		sliceloc.pop(fs_matchpop[i])
		print("i=%s fslice loc: %s"%(i,sliceloc))
		## execute future passive slice at sliceloc
		fslice[i] = xh.junc.pslice(reglist[i], ublocks=[-1], vblocks=range(len(reglist[i].blocks)), **sliceloc)
		## update future passive slice location to true values
		fs_u0[i], fs_v0[i], fs_r0[i]  = 1.*fslice[i].u0, 1.*fslice[i].v0, 1.*fslice[i].r0
	## forward regions
	i = 1*i0 + 1
	while i < len(reglist):
		###### past active slice
		## past active slice input params (not mutually consistent)
		ps_u0[i]  = u0[i] - 0.5*du[i]
		ps_v0[i]  = v0[i] - 0.5*dv[i]
		ps_r0[i]  = 1.*fs_r0[i-1]
		## get past active slice location from inputs and matchpop
		sliceloc = dict(u0=ps_u0[i], v0=ps_v0[i], r0=ps_r0[i])
		sliceloc.pop(ps_matchpop[i])
		print("i=%s pslice loc: %s"%(i,sliceloc))
		## execute past active slice at sliceloc
		pslice[i] = xh.junc.aslice(reglist[i], ublocks=[-1], vblocks=range(len(reglist[i].blocks)), U0=fslice[i-1].U_of_r_at_v0, V0=fslice[i-1].V_of_r_at_u0, r_refs=[fslice[i-1].reg.metfunc.r_ref], **sliceloc)
		## update past active slice location to true values
		ps_u0[i], ps_v0[i], ps_r0[i]  = 1.*pslice[i].u0, 1.*pslice[i].v0, 1.*pslice[i].r0
		#### modify transformations
		## adjust transformations
		reglist[i].U_of_udl = pslice[i].U_of_udl_at_v0
		reglist[i].V_of_vdl = pslice[i].V_of_vdl_at_u0
		#### future passive slice
		## future passive slice input params (not mutually consistent)
		fs_u0[i]  = 1.*ps_u0[i] + 1.*du[i]
		fs_v0[i]  = 1.*ps_v0[i] + 1.*dv[i]
		fs_r0[i]  = 1.*r0f[i]
		## get past active slice location from inputs and matchpop
		sliceloc = dict(u0=fs_u0[i], v0=fs_v0[i], r0=fs_r0[i])
		sliceloc.pop(fs_matchpop[i])
		print("i=%s fslice loc: %s"%(i,sliceloc))
		## execute future passive slice at sliceloc
		fslice[i] = xh.junc.pslice(reglist[i], ublocks=[-1], vblocks=range(len(reglist[i].blocks)), **sliceloc)
		## update future passive slice location to true values
		fs_u0[i], fs_v0[i], fs_r0[i]  = 1.*fslice[i].u0, 1.*fslice[i].v0, 1.*fslice[i].r0
		##### iterate
		## iterate
		i += 1
	## backward regions
	i = 1*i0 - 1
	while i>=0:
		###### future active slice
		## past active slice input params (not mutually consistent)
		fs_u0[i]  = u0[i] - 0.5*du[i]
		fs_v0[i]  = v0[i] - 0.5*dv[i]
		fs_r0[i]  = 1.*ps_r0[i+1]
		## get future active slice location from inputs and matchpop
		sliceloc = dict(u0=fs_u0[i], v0=fs_v0[i], r0=fs_r0[i])
		sliceloc.pop(fs_matchpop[i])
		print("i=%s fslice loc: %s"%(i,sliceloc))
		## execute future active slice at sliceloc
		fslice[i] = xh.junc.aslice(reglist[i], ublocks=[-1], vblocks=range(len(reglist[i].blocks)), U0=pslice[i+1].U_of_r_at_v0, V0=pslice[i+1].V_of_r_at_u0, r_refs=[pslice[i+1].reg.metfunc.r_ref], **sliceloc)
		## update future active slice location to true values
		fs_u0[i], fs_v0[i], fs_r0[i]  = 1.*fslice[i].u0, 1.*fslice[i].v0, 1.*fslice[i].r0
		#### modify transformations
		## adjust transformations
		reglist[i].U_of_udl = fslice[i].U_of_udl_at_v0
		reglist[i].V_of_vdl = fslice[i].V_of_vdl_at_u0
		#### past passive slice
		## past passive slice input params (not mutually consistent)
		ps_u0[i]  = 1.*fs_u0[i] - 1.*du[i]
		ps_v0[i]  = 1.*fs_v0[i] - 1.*dv[i]
		ps_r0[i]  = 1.*r0p[i]
		## get past passive slice location from inputs and matchpop
		sliceloc = dict(u0=ps_u0[i], v0=ps_v0[i], r0=ps_r0[i])
		sliceloc.pop(ps_matchpop[i])
		print("i=%s pslice loc: %s"%(i,sliceloc))
		## execute past passive slice at sliceloc
		pslice[i] = xh.junc.pslice(reglist[i], ublocks=[-1], vblocks=range(len(reglist[i].blocks)), **sliceloc)
		## update future passive slice location to true values
		ps_u0[i], ps_v0[i], ps_r0[i]  = 1.*pslice[i].u0, 1.*pslice[i].v0, 1.*pslice[i].r0
		##### iterate
		## iterate
		i -= 1
	## make sliceparams dict
	chainparams = dict(Rh=1.*np.array(Rh), ps_u0=1.*np.array(ps_u0), ps_v0=1.*np.array(ps_v0), ps_r0=1.*np.array(ps_r0), fs_u0=1.*np.array(fs_u0), fs_v0=1.*np.array(fs_v0), fs_r0=1.*np.array(fs_r0), i0=1*i0, ps_matchmode=ps_matchmode, fs_matchmode=fs_matchmode, funclist=funclist)
	##
	print("\n")
	pprint.pprint(chainparams)
	print("\n")
	## return
	return reglist, chainparams






def chain_masker(reglist, chainparams):
	"""
	"""
	##
	for i in range(len(reglist)):
		## mask interior blocks
		for b in reglist[i].blocks[:-1]:
			## past
			if i>0:
				b.uvbounds.update(dict(vmin=chainparams['ps_v0'][i]))
			## future
			if i<len(reglist)-1:
				b.uvbounds.update(dict(vmax=chainparams['fs_v0'][i]))
		## mask final blocks for part that is always there
		for b in reglist[i].blocks[-1:]:
			## past
			if i>0:
				b.uvbounds.update(dict(vmin=chainparams['ps_v0'][i], umin=chainparams['ps_u0'][i]))
			## future
			if i<len(reglist)-1:
				b.uvbounds.update(dict(vmax=chainparams['fs_v0'][i]))
		## copy final block for parts which depend on radius change values
		for b in reglist[i].blocks[-1:]:
			## copies
			ba = xh.block(b.master, b.j, b.bparams)
			bb = xh.block(b.master, b.j, b.bparams)
			bc = xh.block(b.master, b.j, b.bparams)
			## mask a=top b=bottom c=right
			ba.uvbounds.update(dict(vmin=chainparams['fs_v0'][i], vmax= np.inf, umin=chainparams['ps_u0'][i], umax=chainparams['fs_u0'][i]))
			bb.uvbounds.update(dict(vmin=chainparams['ps_v0'][i], vmax=chainparams['fs_v0'][i], umin=-np.inf, umax=chainparams['ps_u0'][i]))
			bc.uvbounds.update(dict(vmin=chainparams['fs_v0'][i], vmax= np.inf, umin=-np.inf, umax=chainparams['ps_u0'][i]))
		## add bottom if increasing from past
		if i>0 and chainparams['Rh'][i-1]<chainparams['Rh'][i]:
			reglist[i].blocks += [bb]
		## add top if decreasing to future
		if i<len(reglist)-1 and chainparams['Rh'][i+1]<chainparams['Rh'][i]:
			reglist[i].blocks += [ba]
		## add right if both
		if i>0 and i<len(reglist)-1 and chainparams['Rh'][i-1]<chainparams['Rh'][i] and chainparams['Rh'][i+1]<chainparams['Rh'][i]:
			reglist[i].blocks += [bc]
	## add masses to chainparams
	chainparams.update(dict(m=getmm(reglist)))
	## return
	return reglist, chainparams


def shellparams_list(Rmax=1., le=.1, Nevap=5, Tevap=10., Tacc=1., Naccrete=1, functype=xh.mf.schwarzschild, fparams=dict()):
	"""
	"""
	## init
	m, du, dv = xh.evap.SSp.SSduvm(Nevap=1*Nevap, Tevap=1.*Tevap, M=0.5*Rmax, le=1.*le)
	m, du, dv = m[::-1], du[::-1], dv[::-1]
	mdudv = [m, du, dv]
	## get shellparams
	sp = []
	for i in range(len(m)):
		func = functype(R=2.*m[i], **fparams)
		sp += [dict(func=copy.deepcopy(func), Rself=1.*func.fparams['R'], du=1.*du[i], dv=1.*dv[i], le=1.*le, Tevap=1.*Tevap, Nevap=1*Nevap, mdudv=mdudv)]
	## edit final one
	sp[-1]['dv'] = 1.*Tacc/float(max(Naccrete-1,1))
	## print
	pprint.pprint(sp)
	## return
	return sp



def cp_from_fdudv(funclist, du=None, dv=None, le=None, uoff=0., voff=0., ueta=1., veta=1.):
	"""
	"""
	## init
	funclist = funclist
	reglist = [xh.reg.EFreg(funcx, boundary=None, rlines=None) for funcx in funclist]
	Rh = np.array([funclist[i].rj[-2] for i in range(len(funclist))])
	du  = 1.*du
	dv  = 1.*dv
	r0f = 1.*Rh + 1.*le
	r0p = 1.*np.roll(r0f,1)
	u0  = 1.*ueta*np.cumsum(du-du[0]) + 1.*uoff
	v0  = 1.*veta*np.cumsum(dv-dv[0]) + 1.*voff
	ps_matchmode = None #['ru' for i in range(len(funclist))]         ### Edit matchmode here
	fs_matchmode = None #['ru' for i in range(len(funclist))]         ### Edit matchmode here
	## iterator
	ii = range(len(funclist))
	## get rinf
	rinf = np.nan * Rh
	for i in ii:
		ia, ib = max(0, i-1), min(i+2, len(ii))
		rinf[i] = get_rinf_uv0(reglist[ia:ib], v0=1.*v0)
	print(rinf)
	## correct first and last r0 values
	r0p[0]  = 1.*rinf[0]
	r0f[-1] = 1.*rinf[-1]
	## correct r0 values for formation and evaporation
	for i in ii:
		## past
		if i>0:
			## accretion
			if Rh[i]>=Rh[i-1]:
				r0p[i] = 1.*rinf[i]
			## evaporation
			if Rh[i]< Rh[i-1]:
				r0p[i] = 1.*Rh[i-1] + 1.*le
		## future
		if i<len(ii)-1:
			## accretion
			if Rh[i]<=Rh[i+1]:
				r0f[i] = 1.*rinf[i]
			## evaporation
			if Rh[i]> Rh[i+1]:
				r0f[i] = 1.*Rh[i] + 1.*le
	## make cp
	cp = dict(du=1.*du, dv=1.*dv, r0p=1.*r0p, r0f=1.*r0f, u0=1.*u0, v0=1.*v0, ps_matchmode=ps_matchmode, fs_matchmode=fs_matchmode)
	# ## return
	return cp.copy()




def formevap_input(Rmax=1., le=.01, Tevap=1., Tacc=1., Nevap=5, Naccrete=5, uoff=0., voff=0., ueta=1., veta=1., functype0=xh.mf.minkowski, fparams0=dict(), functype1=xh.mf.schwarzschild, fparams1=dict()):
	"""
	Build inputs in reverse order starting from far future.

	funclist, seed=0, du=None, dv=None, r0p=None, r0f=None, u0=None, v0=None, ps_matchmode=None, fs_matchmode=None
	"""
	## init
	funclist = []
	du  = []
	dv  = []
	## final region
	funclist += [functype0(**fparams0)]
	du  += [0.]
	dv  += [0.]
	## evap
	sp = shellparams_list(Rmax=1.*Rmax, Nevap=Nevap, le=1.*le, Tevap=1.*Tevap, Naccrete=1*Naccrete, Tacc=1.*Tacc, functype=functype1, fparams=fparams1)
	for i in range(len(sp)):
		funclist += [sp[i]['func']]
		du  += [sp[i]['du']]
		dv  += [sp[i]['dv']]
	## max radius
	Rmax = sp[-1]['Rself']
	## accrete params
	RR = np.linspace(Rmax,0.5*Rmax, Naccrete)[1:]
	for R in RR:
		funclist += [functype1(R=1.*R, **fparams1)]
		du += [0.]
		dv += [Tacc/float(Naccrete-1)]
	## first region
	funclist += [functype0(**fparams0)]
	du  += [0.]
	dv  += [0.]
	## prep for output
	funclist = funclist[::-1]
	du = np.array(du[::-1])
	dv = np.array(dv[::-1])
	le = 1.*le
	## get chain params
	cp = cp_from_fdudv(funclist, du=1.*du, dv=1.*dv, le=1.*le, uoff=1.*uoff, voff=1.*voff, ueta=1.*ueta, veta=1.*veta)
	##
	pprint.pprint(cp)
	## return
	return funclist, cp



def create_evap(params, seed=0):
	"""
	Takes input parameters of the form:

	"""
	## 
	import pprint
	## print
	pprint.pprint("params = %s"%(params))
	pprint.pprint("seed = %s"%(seed))
	## formevap_input
	print("inputs")
	funclist, cp = xh.evap.formevap_input(**params)
	## funclist_chain
	print("chain")
	reglist, chainparams = xh.evap.funclist_chain(funclist, seed=seed, **cp)
	## chain_masker
	print("mask")
	reglist, chainparams = xh.evap.chain_masker(reglist, chainparams)
	## print
	pprint.pprint(chainparams)
	## return
	return reglist, chainparams


def evapsave(path="temp/temp", params=None, chainparams=None, seed=None, sfp=dict(), temp_only=False, massplot=False):
	"""
	Save figure with timestamp and txt notes.
	"""
	##
	import shutil
	import time
	import pprint
	import matplotlib.pyplot as plt
	## get path with timestamp
	ts = str(time.time()).replace(".","")
	## save figure
	print( "save...")
	plt.figure(1)
	sfpp = dict(dpi=400)
	sfpp.update(sfp)
	plt.savefig("%s_%s.png"%(path,ts), **sfpp)
	print( "save done")
	##save text
	print( "save txt")
	ff = open("%s_%s.txt"%(path,ts), 'w')
	ff.write("%s_%s\n"%(path,ts))
	ff.write('\n')
	ff.write('Input:\nparams=\n%s\nseed=\n%s\n'%(pprint.pformat(params),seed))
	ff.write('\n')
	ff.write('Output:\nchainparams=\n%s\n'%(pprint.pformat(chainparams)))
	ff.close()
	##save massplot
	if massplot==True:
		print( "save massplot...")
		xh.evap.massplot.massplotrc()
		plt.figure(99)
		plt.savefig("%s_%s_mass.png"%(path,ts), **sfpp)
		print( "save done")
	## copy to temp
	print( "copy...")
	## copy normally
	if temp_only==False:
		tempsave = shutil.copy
	if temp_only==True:
		tempsave = shutil.move
	## copy or move
	tempsave("%s_%s.png"%(path,ts), path+"_temp.png")
	tempsave("%s_%s.txt"%(path,ts), path+"_temp.txt")
	tempsave("%s_%s_mass.png"%(path,ts), path+"_temp_mass.png")
	## print
	print( "copy done")




if __name__=='__main__':
	pass








##################################################################################################################



